/*
 *   This is msdos, an msdos/ibmpc emulator.
 *   Copyright (C) 1992  Martin Leisner  leisner@eso.mc.xerox.com 
 * 
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License
 *      as published by the Free Software Foundation; either version 2
 *      of the License, or (at your option) any later version.
 * 
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 * 
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * $Log: 8086engine.c,v $
 * Revision 1.7  1994/07/25  20:13:20  leisner
 * *** empty log message ***
 *
 * Revision 1.6  1993/09/26  23:26:06  leisner
 * changing const?
 *
 * Revision 1.5  1993/01/13  16:35:13  leisner
 * changed effective address lookup
 *
 * Revision 1.4  1993/01/09  21:13:44  leisner
 * *** empty log message ***
 *
 * Revision 1.3  1993/01/03  22:54:05  leisner
 * *** empty log message ***
 *
 */

static char *rcsid = "$Id: 8086engine.c,v 1.7 1994/07/25 20:13:20 leisner Exp leisner $";

#define EMUL_TEST

#include <stdio.h>
#include <setjmp.h>
#include <stdarg.h>
#include <strings.h>
#include <signal.h>
#include <assert.h>


#include "misc.h"
#include "8086regs.h"
#include "8086opcodes1.h"
#include "8086opcodes2.h"
#include "parity.h"
#include "memory.h"
#include "task.h"
#include "dosprotos.h"
#include "magic.h"

#ifdef EMUL_TEST
static jmp_buf after_halt;
#endif

static jmp_buf on_fatal_error;
/* on little-endian machines, this go straight through */

/* ordered like the instructions for fast access */
static WORD_REG *word_reg_table[] = { &regs.ax, &regs.cx, &regs.dx, &regs.bx, 
			              &regs.sp, &regs.bp, &regs.si, &regs.di };
static BYTE_REG *byte_reg_table[] = { &regs.al, &regs.cl, &regs.dl, &regs.bl,
				      &regs.ah, &regs.ch, &regs.dh, &regs.bh };




/* overflow computation when it is saved...we need the two values
 * and the operation 
 * Need distinct values for byte and word operations.
 */
static WORD overflow_word1;
static WORD overflow_word2;	
static BYTE overflow_byte1;
static BYTE overflow_byte2;

typedef enum  foobar { OVERFLOW_ADDITION_BYTE, OVERFLOW_SUBTRACTION_BYTE, 
	       OVERFLOW_ADDITION_WORD, OVERFLOW_SUBTRACTION_WORD,
	       OVERFLOW_DIRECT_SET } OVERFLOW_OP;
static OVERFLOW_OP last_overflow = OVERFLOW_DIRECT_SET;




/* condition flags -- set */
#define BOOL(x) 	((x) ? TRUE : FALSE )
#define SET_CARRY(condition) 	(flags.carry = BOOL(condition))
#define SET_PARITY(condition)	(flags.parity = BOOL(condition))
#ifdef DO_AUXCARRY
#define SET_AUXCARRY(condition) (flags.aux_carry = BOOL(condition) )
#else
#define SET_AUXCARRY(condition)
#endif

#define SET_ZERO(condition)	(flags.zero = BOOL(condition))
#define SET_SIGN(condition)	(flags.sign = BOOL(condition) )

/* condition flags -- test -- this should be efficient -- sizeof should compile out... */
#define TEST_PARITY(result)	(parity_values[result & 0xff])
#define TEST_AUXCARRY(result)	(result & 0x10)	
#define TEST_CARRY(result)	((sizeof(result) == sizeof(int)) ? result & 0x10000 :\
								   result & 0x100)
	
#define TEST_SIGN(result)	((sizeof(result) == sizeof(short)) ? result & 0x8000 : \
								     result & 0x80)
/* effective address -- registers are native */
typedef struct {
	BOOLEAN is_reg;
	void *ptr;
} EA;

#define DUMP_EA(x)	 fprintf(stderr, "is_reg = %x, ptr = %x\n", x.is_reg, x.ptr)

static SEG seg_override = NO_SEG;

/* print regs to files, without a carriage return, in aztec format */
static void print_regs(FILE *pfile, REGS_8086 *regs)
{
	fprintf(pfile, "ax=%04x  bx=%04x  cx=%04x dx=%04x sp=%04x bp=%04x si=%04x di=%04x\n",
                        regs->ax, regs->bx, regs->cx, regs->dx, regs->sp, regs->bp, regs->si, regs->di);
        fprintf(pfile, "cs=%04x ds=%04x ss=%04x es=%04x ip=%04xh",
                        regs->cs, regs->ds, regs->ss, regs->es, regs->ip);
}


static void print_flags(FILE *pfile) 
{					
#define PRINT_FLAG(x)	((x == TRUE) ? '1' : '0')
	fprintf(pfile, "  o=%c d=%c i=%c s=%c z=%c a=%c p=%c c=%c\n",
		PRINT_FLAG(get_overflow()),
		PRINT_FLAG(flags.direction),
		PRINT_FLAG(flags.interrupt_enable),
		PRINT_FLAG(flags.sign),
		PRINT_FLAG(flags.zero),
		PRINT_FLAG(flags.aux_carry),
		PRINT_FLAG(flags.parity),
		PRINT_FLAG(flags.carry));
#undef PRINT_FLAG
}
/* for debugging */
static void print_regs_to_stdout(void)
{
	print_regs(stdout, &regs);
	print_flags(stdout);
}

#define TRACE_CALLS
#ifdef TRACE_CALLS
static FILE *call_trace_file = NULL;
static BOOLEAN do_call_trace = FALSE;

typedef struct {
	WORD from_segment;
	WORD from_offset;
	WORD new_segment;
	WORD new_offset;
} FAR_PTR;


#define SIZE_CALLSTACK 100

static FAR_PTR call_stack[SIZE_CALLSTACK];
static FAR_PTR *pcall_stack = &call_stack[0];

/* this is just for convenience of debugging */
static void reset_call_stack(void)
{
	pcall_stack = &call_stack[0];
	if(call_trace_file)
		do_truncate(call_trace_file);
}

static void add_call_stack(WORD segment, WORD offset)
{
	if(pcall_stack > &call_stack[SIZE_CALLSTACK - 1]) {
		fprintf(stderr, "call stack overflow\n");
		abort();
	}

	pcall_stack->from_offset = regs.ip;
	pcall_stack->from_segment = regs.cs;
	pcall_stack->new_segment = segment;
	pcall_stack->new_offset = offset;
	pcall_stack++;
	show_call_stack();
}

static void sub_call_stack(void)
{
	pcall_stack--;
	
	show_call_stack();
	if(pcall_stack < &call_stack[0]) {
		fprintf(stderr, "call stack underflow\n");
		call_trace("call stack underflow\n");
		pcall_stack = &call_stack[0];
	}
}

static void call_trace(char *fmt, ... )
{
	va_list args;

	if(!call_trace_file) {
		call_trace_file = fopen("call.trace", "w");
		if(!call_trace_file) {
			perror("can't open call trace");
			exit(1);
		}
	}
	va_start(args, fmt);
	vfprintf(call_trace_file, fmt, args);
	fflush(call_trace_file);
	va_end(args);
}

show_call_stack(void)
{
	FAR_PTR *pstack;

	call_trace("call stack after %d instructions:\n", num_instructions);
	for(pstack = pcall_stack - 1; pstack >= &call_stack[0]; pstack--) {
		call_trace("%d -- from %04x:%04x, to %04x:%04x\n", pstack - call_stack,
					pstack->from_segment,  pstack->from_offset,
					pstack->new_segment,  pstack->new_offset);
	}
}

static void trace_call_far(WORD segment, WORD offset)
{
	if(do_call_trace == FALSE) return;

	add_call_stack(segment, offset);
	call_trace("far call to %x:%x\n", segment, offset);
	print_regs(call_trace_file, &regs);
	call_trace("\n");
}

static void trace_call_near(WORD offset)
{
	char *get_symbol(WORD);
	if(do_call_trace == FALSE) return;

	add_call_stack(regs.cs, offset);
	call_trace("near call to %x from %x, stack = %d %s\n",  offset, regs.ip, 
		pcall_stack - &call_stack[0], get_symbol(offset) ); 
	call_trace("\n");
}

static void trace_return_near(WORD from, WORD to)
{
	if(do_call_trace == FALSE) return;

	sub_call_stack();
	call_trace("return near from %x to %x, size = %d\n", from, to,
			pcall_stack - &call_stack[0]);
	call_trace("\n");
}

static void trace_return_far()
{
	if(do_call_trace == FALSE) return;

	sub_call_stack();
	call_trace("return far\n");
	call_trace("\n");
}

#else
static void trace_call_near(WORD offset) { }
static void trace_call_far(WORD segment, WORD offset) { } 
static void trace_return_near() { }
static void trace_return_far() { } 
static void call_trace() { } 
static FILE *call_trace_file;
#endif
#if 1
static int do_debug; 
static int do_trace;
static void  debug_trace( char *fmt, ...)
{
	static FILE *debug_trace_stream;
	 va_list args;

	if(do_debug) {
		if(!debug_trace_stream) {
			debug_trace_stream = fopen("debug.trace", "w");
			if(!debug_trace_stream) {
				perror("Can't open debug.trace");
				exit(1);
			}
		}

		va_start(args, fmt);
		vfprintf(debug_trace_stream, fmt, args);
		
		va_end(args);
	}
}

typedef struct sym_table {
	struct sym_table *next;
	char *name;
	unsigned address;
} SYMBOL;

static SYMBOL *sym_first;	/* symbols in sorted order */
static SYMBOL *sym_last;	/* last symbol -- for adding */

	

void try_to_read_symbol_table(const char *name)
{	
	FILE *symbol_table;
	char symbol_name[80];
	char buffer[80];
	SYMBOL *psym;

	strcpy(symbol_name, name);
	if(rindex(symbol_name, '.')) 
		sprintf(rindex(symbol_name, '.') + 1, "sym");
	else {
		fprintf(stderr, "name has no .?\n");
		return;
	}

	symbol_table = fopen(symbol_name, "r");
	if(!symbol_table)
		return;
	
	/* read in symbol table entries */
	while(fgets(buffer, sizeof buffer, symbol_table) != NULL) {
		psym = malloc(sizeof (SYMBOL));
	
		sscanf(buffer + 6, "%x", &(psym->address));
		psym->name = malloc(strlen(buffer + 9) + 2);
		strcpy(psym->name, buffer + 9);
		*(rindex(psym->name, '\n')) = '\0';
		if(!sym_first) {
			sym_first = psym;
			sym_last = psym;
		} else {
			sym_last->next = psym;
			sym_last = psym;
		}
	}
	fclose(symbol_table);
}


static char *get_symbol(WORD address)
{
	if(sym_first) {
		SYMBOL *psym;


		for(psym = sym_first; psym->next && (psym->next)->address <= address;
				psym = psym->next)
			;

		return psym->name;
	} else  return "no symbols";
}
	
static void print_symbol_table(FILE *file, WORD address)
{
	if(sym_first) {
		SYMBOL *psym;

		for(psym = sym_first; psym->next && (psym->next)->address <= address;
				psym = psym->next)
			;

		fprintf(file, "%s", psym->name);
		if(psym->address != address)
			fprintf(file, "+%d", address - psym->address); 
	} 
	fprintf(file, "\n\n");	
}

static FILE *print_like = NULL;
/* print the regs in a format like aztec C */
void print_like_aztec(void)
{
	extern char *exe_file;

	if(do_trace) {	
		if(seg_override != NO_SEG) 
			return;	/* ignore segment override instructions */

		if(print_like == NULL) {
			char filename[200];

			strcpy(filename, exe_file);
			strcat(filename, ".trc");
			print_like = fopen(filename, "w");
			fprintf(stderr, "making %s\n", filename);
			if(!print_like) {
				perror("can't open write file");
				exit(1);
			}
		}
					
	#define PRINT_FLAG(x)	((x == TRUE) ? '1' : '0')
		print_regs(print_like, &regs);
		fprintf(print_like, "  o=%c d=%c i=%c s=%c z=%c a=%c p=%c c=%c\n",
			PRINT_FLAG(get_overflow()),
			PRINT_FLAG(flags.direction),
			PRINT_FLAG(flags.interrupt_enable),
			PRINT_FLAG(flags.sign),
			PRINT_FLAG(flags.zero),
			PRINT_FLAG(flags.aux_carry),
			PRINT_FLAG(flags.parity),
			PRINT_FLAG(flags.carry));
		print_symbol_table(print_like, regs.ip);		
		fflush(print_like);
	}
}

static void do_truncate(FILE *file)
{
	rewind(file);
	ftruncate(fileno(file), 0);
}

static void rewind_print_like(void)
{
	if(print_like)
		do_truncate(print_like);
}

#else
static int do_trace=0;
void print_like_aztec() { }
static void debug_trace() {} 
void try_to_read_symbol_table(){} 
#endif

#define INSTRUCTION_TRACE /* */

#ifdef INSTRUCTION_TRACE
#define NUM_INSTRUCTIONS 10
/* buffer to save instructions */
static REGS_8086 traced_regs[NUM_INSTRUCTIONS];

/* points to last instruction saved, 
 * we want to dump saved_instruction...NUM_INSTRUCTIONS-1,0, saved_instruction -1
 */
static REGS_8086 *saved_instructions = NULL;

void save_instruction(BYTE byte)
{
	if(!saved_instructions)
		saved_instructions = &traced_regs[0];
	else {
		saved_instructions++;
		if(saved_instructions >= &traced_regs[NUM_INSTRUCTIONS])
			saved_instructions = &traced_regs[0];
		*saved_instructions = regs;
	}
}

#ifndef SAVE_FILE_NAME
#define SAVE_FILE_NAME "saved.instructions"
#endif

void print_saved_instructions(void)
{
	FILE *save_file;
	REGS_8086 *pregs;
	int i;

	if(saved_instructions) {
		pregs = saved_instructions;
		save_file = fopen(SAVE_FILE_NAME, "w");
		if(!save_file) {
			fprintf(stderr, "Can't open %s\n", SAVE_FILE_NAME);
			perror(" " );
			return;
		}
		for(i = 0; i < NUM_INSTRUCTIONS; i++) {
			print_regs(save_file, pregs);
			fprintf(save_file, "\n");
			pregs--;
			if(pregs < &traced_regs[0])
				pregs = &traced_regs[NUM_INSTRUCTIONS - 1];
					
		}
	}
}

#else
inline void save_instruction(BYTE byte)
{
}

inline void print_saved_instructions(void)
{
}
#endif

static void not_yet_implemented(const char *function)
{
	printf("Not implemented %s at %x:%x, instructions = %d\n",
		function, regs.cs, regs.ip, num_instructions);
}


/* 
(Message XNS:1204)
Return-Path: <leisner@vangogh>
Received: from vangogh.wbst147.xerox.com by annapurna.wbst147.xerox.com (4.1/SMI-4.1)
	id AA17257; Sat, 5 Dec 92 16:09:21 EST
Received: by vangogh.wbst147.xerox.com (4.1/SMI-4.0 )
	id AA16491; Sat, 5 Dec 92 16:10:19 EST
Date: Sat, 5 Dec 92 16:10:19 EST
Message-Id: <9212052110.AA16491@vangogh.wbst147.xerox.com>
Postmark: Fri Dec  4 22:09:06 1992 (XNS-SMTP-Gateway:Parc:Xerox)
Sender: dice@praxsys.east.sun:com:Xerox@vangogh
From: dice@praxsys.east.sun:com:Xerox@vangogh
To: leisner:HENR801c:Xerox@vangogh
Subject: Overflow calculations


Hello:  
Well, you're certainly on the right track.  The aux carry
can be safely ignored.  Overflow, while used infrequently, needs
to be emulated correctly.  

One way to proceed with your emulator could be to avoid performing
the overflow condition-code calculation but put in "panic" calls in
the JO,JNO and INTO instructions.  I've never seen anyone check 
the overflow flag via pushf or int.    

Overflow means _signed_ integer overflow.  There are various ways
to calculate it, all being rather expensive in terms of CPU.  
Overflow occurs if the Carry-In of the MSB of the results differs
from the Carry-Out of the the MSB of the result.  

For example ... Addition:
	Overflow occurs if the signs of the addends are the same
	but the sign of the result differs from the sign of the
	addends.  There can be no overflow if the addends have the
	same signs.   

In 16-bit signed 2`s complement:
	Adding 32767 (0x7FFF) + 32 (0x20) yields 32799 (0x801F).
	However there is NO representation of 32799 in 16-bit signed
	representation;  0x801F represents -32737.  

If you're virtualizing a 16-bit ALU op with a 32-bit CPU the following
trick works for _all_ ALU operations:
	int Tmp ;		// 32-bits 
	
	Tmp = (A Op B) ^ A ^ B ; 
	Overflow = (Tmp ^ (Tmp >> 1)) & 0x8000 ; 

In addition to the basic arithmetic operations you'll need to update 
overflow for some of the shift-rotation instructions.  
	
You may want to consider ways to defer or avoid calculating
conditions codes.  In most Intel code streams, only about 1 in 4
ALU ops actually generates a condition code value that is used by
another instruction.  In the other 3 out of 4 times the ICC calculation
is wasted.  Put your compiler hat on and think of it as a dead store (:-). 

Places to look for things:
o	GDE 
	G  = the author's name, which I can't recall.
	DE = DOS emulator
	Check anon-ftp ftp.scri.fsu.edu: pub/hudgens/gde- ....
	That ftp address is from memory so you may to check
	archie or the equivalent for "GDE".  

o	The Motorolla M68020 Users manual has an appendix, 
	"Condition Code Computations" that has a rather 
	lucid explanation.  

o	The old MINIX release used to come with an 8086 emulator
	in the source kit.  The emulator was slow but faithful.  

Enjoy
-	Dave Dice
	
*
*/
static BOOLEAN overflow_addition_word(const WORD arg1, const WORD arg2)
{
	int tmp;

	
	tmp = ( arg1 + arg2) ^ arg1 ^ arg2;
	return BOOL( ((tmp ^ (tmp >> 1)) & 0x8000));
}

static BOOLEAN overflow_addition_byte(const BYTE arg1, const BYTE arg2)
{
	short tmp;


	tmp = ( arg1 + arg2) ^ arg1 ^ arg2;
	return BOOL( ((tmp ^ (tmp >> 1)) & 0x8000));
}

static BOOLEAN overflow_subtraction_word(const WORD arg1, const WORD arg2)
{
	int tmp;

	tmp = ( arg1 - arg2) ^ arg1 ^ arg2;
	return BOOL( ((tmp ^ (tmp >> 1)) & 0x8000));
}
static BOOLEAN overflow_subtraction_byte(const BYTE arg1, const BYTE arg2)
{

	short tmp;


	tmp = ( arg1 -  arg2) ^ arg1 ^ arg2;
	return BOOL( ((tmp ^ (tmp >> 1)) & 0x8000));
}

static inline BOOLEAN compute_overflow(void)
{


	switch(last_overflow) {
		case OVERFLOW_SUBTRACTION_WORD:
			return overflow_subtraction_word(overflow_word1, overflow_word2); 
		case OVERFLOW_SUBTRACTION_BYTE:
			return overflow_subtraction_byte(overflow_byte1, overflow_byte2); 
		case OVERFLOW_ADDITION_WORD:
			return overflow_addition_word(overflow_word1, overflow_word2); 
		case OVERFLOW_ADDITION_BYTE:
			return overflow_addition_byte(overflow_byte1, overflow_byte2); 
		default:
			fatal_error();
	}
}


static BOOLEAN  get_overflow(void)
{
	if(last_overflow != OVERFLOW_DIRECT_SET) {
		flags.overflow = compute_overflow();
		last_overflow = OVERFLOW_DIRECT_SET; 
	}
	return flags.overflow;
}

static void set_overflow_word(OVERFLOW_OP op, WORD value1, WORD value2)
{
	overflow_word1 = value1;
	overflow_word2 = value2;
	last_overflow = op;
}

	
static void set_overflow_byte(OVERFLOW_OP op, BYTE value1, BYTE value2)
{
	overflow_word1 = value1;
	overflow_word2 = value2;
	last_overflow = op;
}

static void set_overflow(BOOLEAN state)
{
	flags.overflow = state;
	last_overflow = OVERFLOW_DIRECT_SET;
}

 
static WORD read_word(const EA ea)
{

	if(ea.is_reg)
		return *(WORD *) ea.ptr;
	else 	return read_pc_word(ea.ptr);
}

static void write_word(const EA ea, const WORD value)
{
	if(ea.is_reg)
		*(WORD *) ea.ptr = value;
	else	write_pc_word((BYTE *) ea.ptr, value);
}


/* return preferred segment if no override, otherwise return the
 * correct override segment 
 */
static BYTE *get_preferred_segment(BYTE *preferred_segment)
{
	switch(seg_override) {
		case NO_SEG:	return preferred_segment;
		case DS_SEG:	return fast_ds;
		case ES_SEG:	return fast_es;
		case SS_SEG:	return fast_ss;
		case CS_SEG:	return fast_cs;
		default:	fatal_error();
	}
}

/* get an instruction out of cs:ip, emulate it and come up with a new CS:IP */
static BYTE fetch(void)
{
	return *(fast_cs + regs.ip++);
}

static WORD fetch_word()
{
	WORD temp;

	temp = fetch();
	return temp  + (fetch() << 8);
}

/* always get the next byte.
 * If the condition is true, ip += next byte
 * Otherwise, throw byte away
 */
static void jump_fetch(const int condition)
{
	char displacement = fetch();	/* is this signed? */

	if(condition)
		regs.ip += displacement;	
}


/* return the top word off stack, advance SP by 2 */
static WORD pop()
{
	WORD temp;

	temp = read_pc_word(fast_ss + regs.sp);
	regs.sp += 2;
#ifdef POP_EXCEPTION
	/* check for sp at 0xffff to generate interrupt 13? */
#endif
	return temp;
}

static void push(const WORD word)
{
#if PUSH_EXCEPTION
#endif

	regs.sp -= 2;
	write_pc_word(fast_ss + regs.sp, word);
}


static void advance_si_word(void)
{
	if(flags.direction == TRUE)
		regs.si -= 2;
	else	regs.si += 2;
}

static void advance_si_byte(void)
{
	if(flags.direction == TRUE)
		regs.si--;
	else	regs.si++;
}

static void advance_di_word(void)
{
	if(flags.direction == TRUE)
		regs.di -= 2;
	else	regs.di += 2;
}

static void advance_di_byte(void)
{
	if(flags.direction == TRUE)
		regs.di--;
	else	regs.di++;
}

/* compute the effecitve address offset based on r_m and mod */
static WORD compute_effective_address_offset(const int mod, const int r_m)
{
	WORD offset;

#ifdef CHECK_EFFECTIVE_ADDRESS
	if(r_m >= 8) {
		fprintf(stderr, "r_m in compute effective address offset is %x\n", r_m);
	}
#endif
	switch (mod) {
	    case 0:
		    if (r_m == 0x6) {
			    /* special case */
			    return fetch_word();
		    }
		    offset = 0;
		    break;
	    case 1:
		    /* sign extended disp low */
		    offset = (char) fetch();
		    break;
	    case 2:
		    offset = fetch_word();
		    break;
	    case 3:
#if 0
		    /* rm is a reg field -- byte or word?? */
		    ea.is_reg = TRUE;
		    ea.ptr = (doing_byte == TRUE) ? 
			     byte_reg_table[r_m] : word_reg_table[r_m];
			
 		   return ea;
#endif
	    default:
		   fatal_error(); 
	}

	switch (r_m) {
	    case 0:		/* EA = (BX) + (SI) + DISP */
		    offset = (int) regs.bx + (int) regs.si + (int) offset;
		    break;
	    case 1:		/* EA = (BX) + (DI) + DISP */
		    offset = (int) regs.bx + (int) regs.di + (int) offset;
		    break;
	    case 2:		/* (BP) + (SI) + DISP */
		    offset = (int) regs.bp + (int) regs.si + (int) offset;
		    break;
	    case 3:		/* (BP) + (DI) + DISP */
		    offset = (int) regs.bp + (int) regs.di + (int) offset;
		    break;
	    case 4:		/* (SI) + DISP       */
		    offset = (int) regs.si + (int) offset;
		    break;
	    case 5:		/* (DI) + DISP		*/
		    offset= (int) regs.di + (int) offset;
		    break;
	    case 6:		/* (BP) + DISP		*/
		    offset = (int) regs.bp + (int) offset;
		    break;
	    case 7:		/* (BX) + DISP		*/
		    offset = (int) regs.bx + (int) offset;
		    break;
	    default:
		    fprintf(stderr, "illegal r/m = %d\n", r_m);
		    fatal_error();
	}

	return offset;


}
/* returns EA either into pc memory or an emulated register...
 * byte or word?
 */
static EA compute_effective_address(const int mod, const int r_m, const BOOLEAN doing_byte)
{
	WORD disp;		/* this is unsigned? */
	void *offset;  /* ?? */
	EA ea;


#ifdef CHECK_EFFECTIVE_ADDRESS
	if(r_m >= 8) {
		fprintf(stderr, "r_m in compute effective address is %x\n", r_m);
	}
#endif

	/* this is a special case -- ea is a register */
	if(mod == 3) {

	    	ea.is_reg = TRUE;
	    	ea.ptr = (doing_byte == TRUE) ? 
	     		byte_reg_table[r_m] : word_reg_table[r_m];
		return ea;	
	}

	ea.is_reg = FALSE;

	if(mod == 0 && r_m == 6) {
		/* special case :  offset relative to ds */
		disp = fetch_word();
		ea.ptr = get_preferred_segment(fast_ds) + disp;
		return ea;
	}
				
	disp = compute_effective_address_offset(mod, r_m);

	switch(r_m) {	
	    case 0:		/* EA = (BX) + (SI) + DISP */
		    offset = get_preferred_segment(fast_ds); 
		    break;
	    case 1:		/* EA = (BX) + (DI) + DISP */
		    offset = get_preferred_segment(fast_ds);
		    break;
	    case 2:		/* (BP) + (SI) + DISP */
		    offset = get_preferred_segment(fast_ss);
		    break;
	    case 3:		/* (BP) + (DI) + DISP */
		    offset = get_preferred_segment(fast_ss);
		    break;
	    case 4:		/* (SI) + DISP       */
		    offset = get_preferred_segment(fast_ds);
		    break;
	    case 5:		/* (DI) + DISP		*/
		    offset = get_preferred_segment(fast_ds);
		    break;
	    case 6:		/* (BP) + DISP		*/
		    offset = get_preferred_segment(fast_ss);
		    break;
	    case 7:		/* (BX) + DISP		*/
		    offset = get_preferred_segment(fast_ds);
		    break;
	    default:
		    fprintf(stderr, "illegal r/m = %d\n", r_m);
		    fatal_error();
	}




        /* looks like with disp = 0xffff and this is a word, an INT13 will be generated? */

	ea.ptr = offset + disp;
#if 0
	debug_trace(" ea.ptr = %x, ea.is_reg = %s ", ea.ptr, (ea.is_reg == TRUE) ? "TRUE" : "FALSE");
	debug_trace(" value = %x\n", read_word(ea));
#endif
	return ea;

}




/* implement LAHF instruction from our flags */
static void lahf_flags(void)
{
	regs.ah = (flags.carry == TRUE) ? 1 : 0;
	regs.ah += ((flags.parity == TRUE) ? 1 : 0) << 2;  
	regs.ah += ((flags.aux_carry == TRUE) ? 1 : 0) << 4;
	regs.ah += ((flags.zero == TRUE) ? 1 : 0) << 6;
	regs.ah += ((flags.sign == TRUE) ? 1 : 0) << 7;
}

/* implement SAHF instruction */
static sahf_flags()
{
	flags.carry = (regs.ah & 1) ? TRUE : FALSE;
	flags.parity = (regs.ah & 4) ? TRUE : FALSE;
	flags.aux_carry = (regs.ah & 0x10) ? TRUE : FALSE;
	flags.zero = (regs.ah & 0x40) ? TRUE : FALSE;
	flags.sign = (regs.ah & 0x80) ? TRUE : FALSE;
}


/* increment register and set flags */
static void increment_reg(WORD_REG *preg)
{
	set_overflow_word(OVERFLOW_ADDITION_WORD, *preg, 1);
	(*preg)++;
	SET_PARITY(TEST_PARITY(*preg));
	SET_ZERO(!*preg);
	SET_SIGN(*preg & 0x8000);
	SET_AUXCARRY(TEST_AUXCARRY(*preg));
}

/* decrement register and set flags */
static void decrement_reg(WORD_REG *preg)
{
	set_overflow_word(OVERFLOW_SUBTRACTION_WORD, *preg, 1);
        (*preg)--;
        SET_PARITY(TEST_PARITY(*preg));
        SET_ZERO(!*preg);
        SET_SIGN(*preg & 0x8000);
        SET_AUXCARRY(TEST_AUXCARRY(*preg));

}

static void do_nop(void)
{
}


/* rotate left, set carry but don't include it in the rotate */
static void do_rol_byte(BYTE *p, int count)
{
	while(count--) {
		if(*p & 0x80) {
			*p <<= 1;
			*p |= 1;
		} else *p <<= 1;
	}
	 
}

static void do_rol_word(WORD *p, int count)
{
        while(count--)  { 
		if(*p & 0x8000) {
			*p <<= 1;
			*p |= 1;
		} else *p <<= 1;
	}
}

static void do_ror_byte(BYTE *p, int count)
{
	while(count--)  {
		if(*p & 1) {
			*p >>= 1;
			*p |= 0x80;
		} else *p >>= 1;
	}
	
}


static void do_ror_word(WORD *p, int count)
{
        while(count--) {
		if(*p & 1) {
			*p >>= 1;
			*p |= 0x8000;
		} else *p >>= 1; 
	} 	
}

static void do_rcr_byte(BYTE *p,  int count)
{
	BOOLEAN  carry;

	while(count--) {
		carry = (*p & 1) ? TRUE : FALSE;
		*p >>= 1;
		if(flags.carry == TRUE)
			*p |= 0x80;
		flags.carry = carry;
	} 
}

static void do_rcr_word(WORD *p, int count)
{

	BOOLEAN  carry;

	while(count--) {
		carry = (*p & 1) ? TRUE : FALSE;
		*p >>= 1;
		if(flags.carry == TRUE)
			*p |= 0x8000;
		flags.carry = carry;
	} 
}

static void do_rcl_byte(BYTE *p, int count)
{
	BOOLEAN  carry;

	while(count--) {
		carry = (*p & 0x80) ? TRUE : FALSE;
		*p <<= 1;
		if(flags.carry == TRUE)
			*p |= 1;
		flags.carry = carry;
	} 
}

static void do_rcl_word(WORD *p, int count)
{
	BOOLEAN  carry;

	while(count--) {
		carry = (*p & 0x8000) ? TRUE : FALSE;
		*p <<= 1;
		if(flags.carry == TRUE)
			*p |= 1;
		flags.carry = carry;
	} 
}

static void do_shl_byte(BYTE *p, int count)
{
	short tmp;

	tmp = *p << count;
	*p = tmp;

	SET_ZERO(!*p);
	SET_SIGN(TEST_SIGN(*p));
	SET_PARITY(TEST_PARITY(*p));

}

static void do_shl_word(WORD *p, int count)
{
	int tmp;

	tmp = *p << count;
	*p = tmp;

	SET_ZERO(!*p);
	SET_SIGN(!*p);
	SET_PARITY(TEST_PARITY(*p));
	SET_CARRY(TEST_CARRY(tmp));
}

static void do_shr_byte(BYTE *p, const int count)
{
	static char shr_carrys[8] = { 1, 2, 4, 8, 0x10, 0x20, 0x40, 0x80 };

	set_overflow(BOOL(*p & 0x80));
	if(count < 8)
		SET_CARRY(BOOL(*p & shr_carrys[count -1]));
	*p >>= count;

	SET_PARITY(TEST_PARITY(*p));
	SET_ZERO(!*p); 
	SET_SIGN(TEST_SIGN(*p));
}


static void do_shr_word(WORD *p, const int count)
{
	static WORD shr_carrys[16] = { 1, 2, 4, 8, 0x10, 0x20, 0x40, 0x080,
					 0x100, 0x200, 0x400, 0x800, 0x1000, 0x2000, 0x4000, 0x8000 };

	set_overflow(BOOL(*p & 0x8000));

	SET_CARRY(BOOL(*p & shr_carrys[count -1 ] ));

	*p >>= count;
	SET_PARITY(TEST_PARITY(*p));
	SET_ZERO(!*p); 
	SET_SIGN(TEST_SIGN(*p));
}

static void do_sar_byte(BYTE *p, const int count)
{
	BOOLEAN sign = BOOL(TEST_SIGN(*p));

	SET_CARRY(BOOL(*p & 1));	
	*p  >>= count;
	if(sign == TRUE)
		 *p |= 0x80;
	SET_PARITY(TEST_PARITY(*p));
	SET_ZERO(!*p); 
	SET_SIGN(TEST_SIGN(*p));
	set_overflow(FALSE);
}

static void do_sar_word(WORD *p, int count)
{
	BOOLEAN sign = BOOL(TEST_SIGN(*p));

	SET_CARRY(BOOL(*p & 1));	

	*p  >>= count;
	if(sign == TRUE)
		 *p |= 0x8000; 

	SET_PARITY(TEST_PARITY(*p));
	SET_ZERO(!*p); 
	SET_SIGN(TEST_SIGN(*p));

	/* only used for single bit shifts -- don't bother testing */
	set_overflow(FALSE);
}

static void bad_shift()
{
	fprintf(stderr, "bad shift\n");
	fatal_error();
	/* illegal opcode ? */
}

/* these are array for the logical byte/word operators */
static void (*logical_byte_ops[8])() = { do_rol_byte, do_ror_byte, do_rcl_byte, do_rcr_byte, 
					 do_shl_byte, do_shr_byte, bad_shift, do_sar_byte };
static void (*logical_word_ops[8])() = { do_rol_word, do_ror_word, do_rcl_word, do_rcr_word, 
					 do_shl_word, do_shr_word, bad_shift, do_sar_word };


static void logical_byte_once()
{
	BYTE secondary_opcode, rm, mod;
        EA effective_address;  
        BYTE secondary_byte;

        get_secondary_byte();
        effective_address = compute_effective_address_byte( mod, rm);

	(*logical_byte_ops[secondary_opcode >> 3])((BYTE *) effective_address.ptr, 1);
}

static void logical_word_once()
{
        BYTE secondary_opcode, rm, mod;
        EA effective_address;  
        BYTE secondary_byte;
	WORD tmp;

        get_secondary_byte();

	
	 effective_address = compute_effective_address_word( mod, rm);	
	 tmp = read_word(effective_address);
	(*logical_word_ops[secondary_opcode >> 3])(&tmp, 1);
	 write_word(effective_address, tmp);
}

static void logical_byte_cl(void)
{
        BYTE secondary_opcode, rm, mod;
        EA effective_address;  
        BYTE secondary_byte;
	int count;

        get_secondary_byte();
        effective_address = compute_effective_address_byte( mod, rm);
	count = regs.cl & 0x1f;

       (*logical_byte_ops[secondary_opcode >> 3])((BYTE *) effective_address.ptr, count);
}

static void logical_word_cl(void)
{
        BYTE secondary_opcode, rm, mod;
        EA effective_address;  
        BYTE secondary_byte;
	int count;
	WORD tmp;

        get_secondary_byte();
	
	effective_address = compute_effective_address_word( mod, rm);
	count = regs.cl & 0x1f;
	tmp = read_word(effective_address);
        (*logical_word_ops[secondary_opcode >> 3])(&tmp, count);
	write_word(effective_address, tmp);
}

static void logical_byte_imm(void)
{
        BYTE secondary_opcode, rm, mod;
        EA effective_address;  
        BYTE secondary_byte;
	int count;
	
        get_secondary_byte();
	
	count = fetch() & 0x1f;
        effective_address = compute_effective_address_byte( mod, rm);

        (*logical_byte_ops[secondary_opcode >> 3])((BYTE *) effective_address.ptr, count);

}

static void logical_word_imm(void)
{
        BYTE secondary_opcode, rm, mod;
        EA effective_address;  
        BYTE secondary_byte;
	int count;
	WORD tmp;

        get_secondary_byte();
	count = fetch() & 0x1f;
	 effective_address = compute_effective_address_word( mod, rm);
	tmp = read_word(effective_address);
        (*logical_word_ops[secondary_opcode >> 3])(&tmp, count);
	write_word(effective_address, tmp);
}


static void do_seg_override(const SEG seg)
{
	if (seg_override != NO_SEG) {
		/* error condition */
		fprintf(stderr, "double seg override\n");
		fatal_error();
	}

	debug_trace("doing segment override for %d\n", (int) seg);	

	seg_override = seg;
	emulate_instruction();
	seg_override = NO_SEG;
}


/* set the flags directly */
static void set_flags_directly(const WORD flag_reg) 
{
#ifdef TEST_FLAG
#error TEST_FLAG used
#endif
#define TEST_FLAG(bit) 	((flag_reg & (1 << bit)) ? TRUE : FALSE)

	flags.carry =  TEST_FLAG(0);	
	flags.parity = TEST_FLAG(2);
	flags.aux_carry = TEST_FLAG(4);
	flags.zero = TEST_FLAG(6);
	flags.sign = TEST_FLAG(7);
	set_overflow(TEST_FLAG(11));
	flags.trap = TEST_FLAG(8);
	flags.interrupt_enable = TEST_FLAG(9);
	flags.direction = TEST_FLAG(10);
	
#undef TEST_FLAG
}


static WORD get_flags_directly(void) 
{
	WORD temp = 0;


	if(flags.carry == TRUE) temp |= 1;
	if(flags.parity == TRUE) temp |= (1 << 2);
	if(flags.aux_carry == TRUE) temp |= (1 << 4);
	if(flags.zero == TRUE) temp |= (1 << 6);
	if(flags.sign == TRUE) temp |= (1 << 7);
	if(get_overflow() == TRUE) temp |= (1 << 11);
	if(flags.trap == TRUE) temp |= (1 << 8);
	return temp;
	
}

static void not_byte(BYTE *p)
{
	*p = ~*p; /* ? */
	/* NO flags */
}

static void not_word(WORD *p)
{
	*p = ~*p;
	/* no flags */
}

static void neg_byte(BYTE *p)
{
	char result;

	result = -*p;
	
	SET_CARRY(*p != 0);	
	SET_PARITY(TEST_PARITY(result));	
	SET_SIGN(TEST_SIGN(result));
	set_overflow_byte(OVERFLOW_SUBTRACTION_BYTE, 0, *p);	
	SET_ZERO(!result);
	*p = result;
	
}

static void neg_word(WORD *p)
{
	int result;

	result = -*p;
	
	set_overflow_word(OVERFLOW_SUBTRACTION_WORD, 0, *p);	
	SET_CARRY(*p != 0);	
	SET_PARITY(TEST_PARITY(result));	
	SET_SIGN(TEST_SIGN(result));
	SET_ZERO(!result);
	*p = result;
}

static void mul_byte(const BYTE p)
{
	regs.ax = (unsigned char)regs.al &  (unsigned char) p;
	SET_CARRY(BOOL(regs.ah));
	set_overflow(BOOL(regs.ah));
}

static void mul_word(const WORD p)
{
	unsigned temp;

	temp = (unsigned short ) regs.ax * (unsigned short)p;
	regs.ax = temp & 0xffff;
	regs.dx = temp >> 16;
	SET_CARRY(BOOL(regs.dx));
	set_overflow(BOOL(regs.dx));
	
}

static void imul_byte(const BYTE p)
{
	regs.ax = (char) regs.al * (char) p;

}

static void imul_word(const WORD p)
{
	int temp;

	temp = (short) regs.ax * (short) p;
	regs.dx = temp >> 16;
	regs.ax = temp & 0xffff;
}

/* opcode of form IMUL rw,db */
static inline void do_imul_imm_byte(void )
{
	BYTE secondary_opcode, rm, mod;
	char byte;
	BYTE secondary_byte;
	EA effective_address;
	
	get_secondary_byte();
	effective_address = compute_effective_address_word(mod, rm);	
	byte = fetch();	
	
	(short) *word_reg_table[secondary_opcode >> 3] = read_word(effective_address) * byte;
}

static inline void do_imul_imm_word(void)
{
	BYTE secondary_byte;
	BYTE secondary_opcode, rm, mod;
	EA effective_address;
	char  byte;
	short ea_word;	

	get_secondary_byte();

	effective_address = compute_effective_address_word(mod, rm);
	byte = fetch();	
	ea_word = read_word(effective_address);
	*word_reg_table[secondary_opcode >> 3] = ea_word * byte;
}

/* unsigned divide -- div */
static void div_byte(const BYTE p)
{
	unsigned short tmp;
	
	tmp = regs.ax;

	regs.al = tmp/p;
	regs.ah = tmp % p;	
}

static void div_word(const WORD p)
{
	unsigned tmp;
	
	tmp = (regs.dx << 16) + regs.ax;

	regs.ax = tmp/p;
	regs.dx = tmp % p;
}

static void idiv_byte(const BYTE p)
{
	short tmp;
	
	tmp = regs.ax;

	regs.al = tmp/(char ) p;
	regs.ah = tmp % (char ) p;	
 
}

static void idiv_word(const WORD p)
{
	int tmp;
	
	tmp = (regs.dx << 16) + regs.ax;

	regs.ax = tmp/(short )p;
	regs.dx = tmp % (short)p;

}
	
/*
 * Opcodes are:
 * 	TEST
 * 	NOT
 * 	NEG
 * 	MUL
 * 	IMUL
 * 	DIV
 * 	IDIV
 */
static void do_f6_byte_opcodes(void)
{
	BYTE secondary_byte;
	BYTE secondary_opcode, rm, mod;
	EA effective_address;

	get_secondary_byte();
	
	effective_address = compute_effective_address_byte( mod, rm);
 
	switch(secondary_opcode) {
		case TEST:	test_byte_noptr(* (BYTE *) effective_address.ptr,  fetch()); break;	
		case NOT:	not_byte(effective_address.ptr); break;
		case NEG:	neg_byte(effective_address.ptr); break;
		case MUL:	mul_byte(*(BYTE *)effective_address.ptr); break;
		case IMUL:	imul_byte(*(BYTE *)effective_address.ptr); break;
		case DIV:	div_byte(*(BYTE *)effective_address.ptr); break;
		case IDIV:	idiv_byte(*(BYTE *) effective_address.ptr); break;
		default:	
			fatal_error();
	} 
}

static void do_f7_word_opcodes(void)
{
        BYTE secondary_byte;
        BYTE secondary_opcode, rm, mod;
        EA effective_address;
	WORD tmp;
        get_secondary_byte();

        effective_address = compute_effective_address_word( mod, rm);
	tmp = read_word(effective_address);
	switch(secondary_opcode) {
		case TEST:	test_word_noptr(tmp,  fetch_word()); break;	
		case NOT:	not_word(&tmp); write_word(effective_address, tmp); break;
		case NEG:	neg_word(&tmp); write_word(effective_address, tmp); break;
		case MUL:	mul_word(tmp); break;
		case IMUL:	imul_word(tmp); break;
		case DIV:	div_word(tmp); break;
		case IDIV:	idiv_word(tmp); break;
		default:	
			fatal_error();
	}


}


static void increment_ea_byte(unsigned char *effective_address)
{
	unsigned short temp;

	set_overflow_byte(OVERFLOW_ADDITION_BYTE, *effective_address, 1);
        
	temp = ((unsigned short) *effective_address ) + 1;
        *effective_address = temp;        /* will implement roundoff? */


        SET_ZERO(!*effective_address);
        SET_SIGN(TEST_SIGN(*effective_address));
        SET_AUXCARRY(TEST_AUXCARRY(temp));
        SET_PARITY(TEST_PARITY(temp));

}

static void increment_ea_word(const EA effective_address)
{
	WORD temp;


	temp = read_word(effective_address);
	increment_reg(&temp);
	write_word(effective_address, temp);

}

static void decrement_ea_word(const EA effective_address)
{
        WORD temp;
	
	temp = read_word(effective_address);
	decrement_reg(&temp);
	write_word(effective_address, temp);



}

static void decrement_ea_byte(char *effective_address)
{
	unsigned short temp;


	set_overflow_byte(OVERFLOW_SUBTRACTION_BYTE,  *effective_address, 1);	
	temp = ((unsigned short) *effective_address );
	temp--;
	*effective_address = temp;	 /* will implement roundoff? */

        SET_ZERO(!temp);
        SET_SIGN(TEST_SIGN(*effective_address) );
        SET_AUXCARRY(TEST_AUXCARRY(temp));
        SET_PARITY(TEST_PARITY(temp));
        
 
}


/* only have increment/decrement byte here */
static void do_fe_byte_opcodes(void) 
{

        BYTE secondary_byte;
        BYTE secondary_opcode, rm, mod;
	EA effective_address;

	get_secondary_byte();
	effective_address = compute_effective_address_byte(mod, rm);

        switch(secondary_opcode) {
		case INC_EA:
			increment_ea_byte((BYTE *) effective_address.ptr);
			break;
		case DEC_EA:
			decrement_ea_byte((BYTE *) effective_address.ptr);
			break;
		default:
			fprintf(stderr, "Primary opcode = fe, secondary is %x\n", secondary_opcode);
			fatal_error();

			break;
	}
}

static void do_call_near(const WORD newoffset)
{
	push(regs.ip);
	trace_call_near(regs.ip + newoffset);
	regs.ip += newoffset;
}

static void do_call_far(const WORD newsegment, const WORD newoffset)
{
	trace_call_far(newsegment, newoffset);
	push(regs.cs);
	push(regs.ip);
	set_selector(CS_SEG, newsegment);
	regs.ip = newoffset;
}

static void do_call_near_imm(void)
{
	do_call_near(fetch_word());
}

static void do_call_far_imm(void)
{
	WORD segment, offset;

	offset = fetch_word();
	segment = fetch_word();

	do_call_far(segment, offset);
}

static void push_ea_word(EA effective_address)
{
	if(effective_address.is_reg == TRUE)
		push(effective_address.ptr);
	else	push(read_pc_word(effective_address.ptr));	/* flipping bytes twice...
								 * may be an optimization
								 */
}

/* try to get the next two bytes from an EA (must be non-register)
 * and return the word
 */
static WORD far_ptr_segment(EA effective_address)
{
	if(effective_address.is_reg == TRUE) {
		fatal_error();
	} else return read_pc_word(effective_address.ptr + 2);
}

		
static void do_ff_word_opcodes()
{
        BYTE secondary_byte;
        BYTE secondary_opcode, rm, mod;
	EA effective_address;
	WORD newsegment, newoffset; 	/* for far jmps/calls */

	get_secondary_byte();
	effective_address = compute_effective_address_word(mod, rm);

        switch(secondary_opcode) {
		case INC_EA:
			increment_ea_word(effective_address);
			break;
		case DEC_EA:
			decrement_ea_word(effective_address);
			break;
		case PUSH_EA:
			push_ea_word(effective_address);
			break;
		case JMP_NEAR_EA:
			regs.ip = read_word(effective_address);
			break;
		case JMP_FAR_EA:
			newoffset = read_word(effective_address);
			newsegment = far_ptr_segment(effective_address);
			regs.ip = newoffset;
			set_selector(CS_SEG, newsegment);
			break;
		case CALL_NEAR_EA:
			trace_call_near(read_word(effective_address));
			push(regs.ip);
			regs.ip = read_word(effective_address);
			break;
		case CALL_FAR_EA:
			newoffset = read_word(effective_address);
			newsegment = far_ptr_segment(effective_address);
			do_call_far(newsegment, newoffset);					
			break;
		default:
                        fatal_error();
        }

}




/* low level routines to or/and/xor/test/xchg/mov bytes and words.
 * sets the approriate flags.
 * Note all bytes can point anywhere (either internal or in emulated pc ram).
 * Words have to be native, the caller is responsible to do the byte order reversals to pc ram
 */
static void or_byte(BYTE * const dest, BYTE * const src)
{
	*dest |= *src;
	SET_CARRY(0);
	SET_PARITY(TEST_PARITY(*dest));
	set_overflow(FALSE);
	SET_SIGN(TEST_SIGN(*dest));
	SET_ZERO(!*dest);
}

static void or_word(WORD *const dest, WORD *const src)
{
	*dest |= *src;
        SET_CARRY(0);
        SET_PARITY(TEST_PARITY(*dest));
	set_overflow(FALSE);
        SET_SIGN(TEST_SIGN(*dest));
        SET_ZERO(!*dest);
	
}

static void and_byte(BYTE *const dest, const BYTE *const src)
{
        *dest &= *src;
        SET_CARRY(0);
        SET_PARITY(TEST_PARITY(*dest));
	set_overflow(FALSE);
        SET_SIGN(TEST_SIGN(*dest));
        SET_ZERO(!*dest);
}

static void and_word(WORD *const dest, const WORD *const src)
{
        *dest &= *src;   
        SET_CARRY(0);  
        SET_PARITY(TEST_PARITY(*dest));
	set_overflow(FALSE);
        SET_SIGN(TEST_SIGN(*dest));
        SET_ZERO(!*dest);

}
static void xor_byte(BYTE *const dest, const BYTE *const src)
{
        *dest ^= *src;
        SET_CARRY(0);
        SET_PARITY(TEST_PARITY(*dest));
	set_overflow(FALSE);
        SET_SIGN(TEST_SIGN(*dest));
        SET_ZERO(!*dest);
}

static void xor_word(WORD *const dest, const WORD *const src)
{
        *dest ^= *src;   
        SET_CARRY(0);  
        SET_PARITY(TEST_PARITY(*dest));
	set_overflow(FALSE);
        SET_SIGN(TEST_SIGN(*dest));
        SET_ZERO(!*dest);

}

static void xchg_byte(BYTE *const first, BYTE *const second)
{
	BYTE tmp = *first;
	
	*first = *second;
	*second = tmp;
}

static void xchg_word(WORD *const first, WORD *const second)
{
        WORD tmp = *first;
        
        *first = *second;
        *second = tmp;
}

static void test_byte(const BYTE *const first, const BYTE *const second)
{
	test_byte_noptr(*first, *second);
}

static void test_byte_noptr(const BYTE first, const BYTE second)
{
	BYTE result;

	result = first & second;
	SET_CARRY(0);
	set_overflow(FALSE);
	SET_SIGN(TEST_SIGN(result));
	SET_ZERO(!result);
	SET_PARITY(TEST_PARITY(result));
}

static void test_word_noptr(const WORD first, const WORD second)
{
        WORD result;

        result = first & second;
        SET_CARRY(0);
	set_overflow(FALSE);
        SET_SIGN(TEST_SIGN(result));
        SET_ZERO(!result);
        SET_PARITY(TEST_PARITY(result));

}

static void test_word(const WORD *const first, const WORD *const second)
{
	test_word_noptr(*first, *second);
}


static void add_byte(BYTE *dest, BYTE *src)
{
	short result;

	set_overflow_byte(OVERFLOW_ADDITION_BYTE, *dest, *src);	
	result = (short) *dest + (short) *src;
	*dest = result;
	SET_CARRY(TEST_CARRY(result));
	SET_SIGN(TEST_SIGN(*dest));
	SET_AUXCARRY(TEST_AUXCARRY(*dest));
	SET_ZERO(!*dest);
}

static void add_word(WORD *dest, WORD *src)
{
        int result;

	set_overflow_word(OVERFLOW_ADDITION_WORD, *dest, *src);        
        result = (int) *dest + (int) *src;
        *dest = result;
        SET_CARRY(TEST_CARRY(result));
        SET_SIGN(TEST_SIGN(*dest));     
        SET_AUXCARRY(TEST_AUXCARRY(*dest));    
        SET_ZERO(!*dest);  
	SET_PARITY(TEST_PARITY(result));   
}

static void adc_byte(BYTE *dest, BYTE *src)
{
        short result;
	BYTE with_carry;

	with_carry = *src + ((flags.carry == TRUE) ? 1 : 0);
	
	set_overflow_byte(OVERFLOW_ADDITION_BYTE, *dest, with_carry);	
        
        result = (short) *dest + (short) with_carry;
        *dest = result;
        SET_CARRY(TEST_CARRY(result));
        SET_SIGN(TEST_SIGN(*dest));     
        SET_AUXCARRY(TEST_AUXCARRY(*dest));    
        SET_ZERO(!*dest);
	SET_PARITY(TEST_PARITY(*dest));      
}

static void adc_word(WORD *dest, WORD *src)
{
        int result;
	WORD  with_carry;

	with_carry = *src + ((flags.carry == TRUE) ? 1 : 0);
	
	set_overflow_byte(OVERFLOW_ADDITION_BYTE, *dest, with_carry);	
        result = (int) *dest + (int) with_carry;
        *dest = result;
        SET_CARRY(TEST_CARRY(result));
        SET_SIGN(TEST_SIGN(*dest));     
        SET_AUXCARRY(TEST_AUXCARRY(*dest));    
        SET_ZERO(!*dest);      
	SET_PARITY(TEST_PARITY(*dest));
}

/* set the subtraction flags (and return the answer) -- this is
 * used for the sub and cmp instructions
 */
static WORD set_sub_word_flags(const WORD arg1, const WORD  arg2)
{
	int result;

	result = (int) arg1 - (int) arg2;
	set_overflow_word(OVERFLOW_SUBTRACTION_WORD, arg1, arg2);
	SET_PARITY(TEST_PARITY(result));
	SET_SIGN(BOOL(result & 0x8000));
	SET_CARRY(TEST_CARRY(result));
	SET_ZERO(arg1 == arg2);	
	
	return (WORD) result;
}

static BYTE set_sub_byte_flags(const BYTE arg1, const BYTE arg2)
{
	short result;

	result = (short) arg1 - (short) arg2;
	set_overflow_byte(OVERFLOW_SUBTRACTION_BYTE, arg1, arg2);

	SET_PARITY(TEST_PARITY(result));
	SET_SIGN(BOOL(result & 0x80));
	SET_CARRY(TEST_CARRY(result));
	SET_ZERO(arg1 == arg2);	
	return (BYTE) result;
}

static void sub_byte(BYTE *dest, BYTE *src)
{
	*dest =	set_sub_byte_flags(*dest, *src);

}

static void sub_word(WORD *dest, WORD *src)
{
	*dest = set_sub_word_flags(*dest, *src);
}

static void sbb_byte(BYTE *dest, BYTE *src)
{
	*dest = set_sub_byte_flags(*dest,
				   *src + ((flags.carry == TRUE) ? 1 : 0));
}

static void sbb_word(WORD *dest, WORD *src)
{
	*dest = set_sub_word_flags(*dest,
				   *src + ((flags.carry == TRUE) ? 1 : 0));
}

static void cmp_byte_noptr(const BYTE op1, const BYTE op2)
{
	set_sub_byte_flags(op1, op2);
}

static void cmp_byte(BYTE *op1, BYTE *op2)
{
	cmp_byte_noptr(*op1, *op2);
}

static void cmp_word_noptr(const WORD op1, const WORD op2)
{
	set_sub_word_flags(op1, op2);
}


static void cmp_word(WORD *op1, WORD *op2)
{
	cmp_word_noptr(*op1, *op2);
	
}


static void mov_byte(BYTE *dest, BYTE *src)
{
	*dest = *src;
}

static void mov_word(WORD *dest, WORD *src)
{
	*dest = *src;
}
	
static void reg_to_eabyte(void (*const func)(BYTE *, BYTE *))
{
	BYTE secondary_byte;
        BYTE secondary_opcode, rm, mod;
        EA effective_address;

        get_secondary_byte();
        effective_address = compute_effective_address_byte(mod, rm);

	(*func)((BYTE *) effective_address.ptr, byte_reg_table[secondary_opcode >> 3]);

}

static void reg_to_eaword(void (*const func)(WORD *, WORD *))
{
        BYTE secondary_byte;
        BYTE secondary_opcode, rm, mod;
        EA effective_address;
	WORD tmp;

        get_secondary_byte();
        effective_address = compute_effective_address_word(mod, rm);
	
	tmp = read_word(effective_address);
	
        (*func)(&tmp, word_reg_table[secondary_opcode >> 3]);
	
	write_word(effective_address, tmp);
}

static void eabyte_to_reg(void (*const func)(BYTE *, BYTE *))
{
        BYTE secondary_byte;
        BYTE secondary_opcode, rm, mod;
        EA effective_address;

        get_secondary_byte();
        effective_address = compute_effective_address_byte(mod, rm);

        (*func)(byte_reg_table[secondary_opcode >> 3], (BYTE *) effective_address.ptr);

}

static void eaword_to_reg(void (*const func)(WORD *, WORD *))
{
        BYTE secondary_byte;
        BYTE secondary_opcode, rm, mod;
        EA effective_address;
	WORD tmp;

        get_secondary_byte();
        effective_address = compute_effective_address_word(mod, rm);
	tmp = read_word( effective_address );
        (*func)(word_reg_table[secondary_opcode >> 3], &tmp);

}

static void imm_to_al_byte(void (*const func)(BYTE *, BYTE *))
{
	BYTE imm;

	imm = fetch();

	(*func)(&regs.al, &imm);
}

static void imm_to_ax_word(void (*const func)(WORD *, WORD *))
{
	WORD imm;

	imm = fetch_word();
	
	(*func)(&regs.ax, &imm);
}

static void imm_to_eabyte(void (*const func)(BYTE *, BYTE *))
{
        BYTE secondary_byte;
        BYTE secondary_opcode, rm, mod;
        EA effective_address;
	BYTE data;

        get_secondary_byte();
        effective_address = compute_effective_address_byte(mod, rm);
	data = fetch();

	(*func)((BYTE *) effective_address.ptr, &data);
}

static void imm_to_eaword(void (*const func)(WORD *, WORD *))
{
        BYTE secondary_byte;
        BYTE secondary_opcode, rm, mod;
        EA effective_address;
	WORD data;

        get_secondary_byte();
        effective_address = compute_effective_address_word(mod, rm);
	data = fetch_word();

	(*func)((WORD *) effective_address.ptr, &data);
}
	
static void return_far()
{
	trace_return_far();
	regs.ip = pop();
	set_selector(CS_SEG, pop());
}

static void return_near()
{
	WORD tmp = regs.ip;

	regs.ip = pop();	
	trace_return_near(tmp, regs.ip);
}

static void return_pop_words(BYTE bytes)
{
	int words  = bytes/2;
	while(words--)
		pop();
}

static void return_near_pop()
{
	BYTE words = fetch();	
	return_near();
	return_pop_words(words);
}

static void return_far_pop()
{
	BYTE words = fetch();
	return_far();
	return_pop_words(words);

}



static void do_80_byte_opcodes()
{
        BYTE secondary_byte;
        BYTE secondary_opcode, rm, mod;
        EA effective_address;
	BYTE data;

	
        get_secondary_byte();
        effective_address = compute_effective_address_byte(mod, rm);

	data = fetch();
	
	switch(secondary_opcode) {
		case AND:	and_byte(effective_address.ptr, &data);	break;
		case OR:	or_byte(effective_address.ptr, &data);break;
		case ADC:	adc_byte(effective_address.ptr, &data);break;
		case ADD:	add_byte(effective_address.ptr, &data); break;
		case SUB: 	sub_byte(effective_address.ptr, &data);	break;
		case SBB:	sbb_byte(effective_address.ptr,  &data);	break;
		case XOR:	xor_byte(effective_address.ptr, &data); break;
		case CMP: 	cmp_byte(effective_address.ptr, &data);	break;
		default:
			fatal_error();
	}

}

static void do_81_word_opcodes()
{
        BYTE secondary_byte;
        BYTE secondary_opcode, rm, mod;
        EA effective_address;
	WORD ea_data;
	WORD data;

        get_secondary_byte();
        effective_address = compute_effective_address_word(mod, rm);
	data = fetch_word();
	ea_data = read_word(effective_address);

        switch(secondary_opcode) {
		case AND:	and_word(&ea_data, &data);	break;
		case OR:	or_word(&ea_data, &data);break;
		case ADC:	adc_word(&ea_data, &data);break;
		case ADD:	add_word(&ea_data, &data); break;
		case SUB: 	sub_word(&ea_data, &data);	break;
		case SBB:	sbb_word(&ea_data, &data);	break;
		case XOR:	xor_word(&ea_data, &data); break;
		case CMP:	cmp_word(&ea_data, &data);	break;
		default:	fatal_error();	

        }
	/* isn't necessary for cmp, but don't want to do another test */
	write_word(effective_address,  ea_data);	
}

static void do_83_byte_into_word_opcodes(void)
{
        BYTE secondary_byte;
        BYTE secondary_opcode, rm, mod;
        EA effective_address;
	WORD data;
	WORD ea_data;

        get_secondary_byte();
        effective_address = compute_effective_address_word(mod, rm);
	ea_data = read_word(effective_address);

	data = (char) fetch();	/* sign extend */

	switch(secondary_opcode) {
		case ADD:	add_word(&ea_data, &data); break;
		case ADC:	adc_word(&ea_data, &data); break;
		case SUB:	sub_word(&ea_data, &data); break;
		case SBB:	sbb_word(&ea_data, &data); break;
		case CMP:	cmp_word(&ea_data, &data); break;
		/* this is very strange -- not documented, but
		 * msh uses it 
		 */
		case AND:	and_word(&ea_data, &data); break;
		/* another strange one -- not documented? */
		case OR:	or_word(&ea_data, &data); break;
		default:
			fatal_error();
	}
	/* isn't necessary for cmp, but don't want to do another test */
	write_word(effective_address, ea_data);
}

static void mov_to_eabyte()
{
        BYTE secondary_byte;
        BYTE secondary_opcode, rm, mod;
        EA effective_address;

        get_secondary_byte();
        effective_address = compute_effective_address_byte(mod, rm);

	*(BYTE *) effective_address.ptr = *byte_reg_table[secondary_opcode >> 3];
	
}

static void mov_imm_byte_to_ea()
{
        BYTE secondary_byte;
        BYTE secondary_opcode, rm, mod;
        EA effective_address;

        get_secondary_byte();
        effective_address = compute_effective_address_byte(mod, rm);

        *(BYTE *) effective_address.ptr = fetch();
        

}

static void mov_imm_word_to_ea(void)
{

        BYTE secondary_byte;
        BYTE secondary_opcode, rm, mod;
        EA effective_address;

        get_secondary_byte();
        effective_address = compute_effective_address_byte(mod, rm);

	write_word(effective_address, fetch_word());

}

static void mov_to_eaword()
{
        BYTE secondary_byte;
        BYTE secondary_opcode, rm, mod;
        EA effective_address;

        get_secondary_byte();
        effective_address = compute_effective_address_word(mod, rm);

	write_word(effective_address,  *word_reg_table[secondary_opcode >> 3]);
}

static void mov_to_wordreg()
{
        BYTE secondary_byte;
        BYTE secondary_opcode, rm, mod;
        EA effective_address;
        get_secondary_byte();
        effective_address = compute_effective_address_word(mod, rm);

	*word_reg_table[secondary_opcode >> 3] = read_word(effective_address);

}

static void mov_to_bytereg()
{
        BYTE secondary_byte;
        BYTE secondary_opcode, rm, mod;
        EA effective_address;

        get_secondary_byte();
        effective_address = compute_effective_address_byte(mod, rm);

	*byte_reg_table[secondary_opcode >> 3] = *(BYTE *) effective_address.ptr;
}


static void mov_seg_to_ea()
{
        BYTE secondary_byte;
        BYTE secondary_opcode, rm, mod;
        EA effective_address;

        get_secondary_byte();
        effective_address = compute_effective_address_word(mod, rm);

	switch(secondary_opcode) {
		case ES:	write_word(effective_address, regs.es); break;
		case DS:	write_word(effective_address, regs.ds); break;
		case CS:	write_word(effective_address, regs.cs); break;
		case SS:	write_word(effective_address, regs.ss); break;
		default:
			fatal_error();
	}
}

static void mov_ea_to_seg()
{
        BYTE secondary_byte;
        BYTE secondary_opcode, rm, mod;
        EA effective_address;
	WORD data;

        get_secondary_byte();
        effective_address = compute_effective_address_word(mod, rm);
	
	data = read_word(effective_address);

	switch(secondary_opcode) {
		case DS:	set_selector(DS_SEG, data); break;
		case ES:	set_selector(ES_SEG, data); break;
		case SS:	set_selector(SS_SEG, data); break;
		default:
			fatal_error();
	}
}



static void do_movs_byte(void)
{
	*(BYTE *) (fast_es + regs.di) = *(BYTE *)( get_preferred_segment(fast_ds) + regs.si);
	advance_si_byte();
	advance_di_byte();

}

static void do_movs_word(void)
{
	write_pc_word(fast_es + regs.di , read_pc_word(get_preferred_segment(fast_ds) + regs.si));
	advance_si_word();
	advance_di_word();
}

static void do_scas_byte(void)
{
	cmp_byte(&regs.al, (BYTE *) fast_es + regs.di);
	advance_di_byte();
}

static void do_stos_byte(void)
{
	*(BYTE *) (fast_es + regs.di) = regs.al;
	advance_di_byte();
}

static void do_stos_word(void)
{
	write_pc_word(fast_es + regs.di, regs.ax);
	advance_di_word();
}

static void do_lods_byte(void)
{
	regs.al = *(BYTE *) (fast_ds + regs.si);
	advance_si_byte();
}

static void do_lods_word(void)
{
	regs.ax = read_pc_word(fast_ds + regs.si);
	advance_si_word();	
}
	
static void do_scas_word(void)
{
	cmp_word_noptr(regs.ax, read_pc_word(fast_es + regs.di));
	advance_di_word();

}

static void do_cmps_byte(void)
{

	cmp_byte(get_preferred_segment(fast_ds) + regs.si, fast_es + regs.di);
	advance_si_byte();
	advance_di_byte();
}

static void do_cmps_word(void)
{

	cmp_word_noptr(read_pc_word(get_preferred_segment(fast_ds) + regs.si), read_pc_word(fast_es + regs.di));


	advance_si_word();
	advance_di_word();
}

static BYTE in_byte(const WORD port)
{
	fatal_error();
}

static WORD in_word(const WORD port)
{
	fatal_error();
}

static void out_word(const WORD port, const WORD value )
{
	fatal_error();
}

static void out_byte(WORD port, BYTE value)
{
	fatal_error();
}

static void do_ins_byte()
{
	*(BYTE *) (fast_es + regs.di) = in_byte(regs.dx);
	advance_di_byte();

}

static void do_ins_word()
{
	write_pc_word(fast_es + regs.di, in_word(regs.dx));
	advance_di_word();

}

static void do_outs_byte()
{
	out_byte(regs.dx, *(BYTE *) (fast_es + regs.di));
	advance_di_byte();

}

static void do_outs_word()
{
	out_word(regs.dx, read_pc_word(fast_es + regs.di));
	advance_di_word();
}

/* Returns TRUE if opcode is one of the REPNE (SCAS, CMPS).
 * REP also look here about whether to check the Z flag...
 */
static BOOLEAN check_repne_opcode(const BYTE opcode)
{
	switch(opcode) {
		case SCAS_BYTE:
		case SCAS_WORD:
		case CMPS_BYTE:
		case CMPS_WORD:
			return TRUE;
		default:	
			return FALSE;
	}	
}	

/* only process rep opcodes which don't check the zero flag */
static void do_only_rep_opcodes(const BYTE opcode)
{

	while(regs.cx) {
		switch(opcode) {
			case INS_BYTE:	do_ins_byte(); break;
			case INS_WORD:	do_ins_word(); break;
			case MOVS_BYTE:	do_movs_byte(); break;
			case MOVS_WORD:	do_movs_word(); break;
			case OUTS_BYTE:	do_outs_byte(); break;
			case OUTS_WORD:	do_outs_word(); break;
			case STOS_BYTE:	do_stos_byte(); break;
			case STOS_WORD: do_stos_word(); break;
			default:
				fprintf(stderr, "invalid opcode %x in REP\n", opcode);
				fatal_error();
		}
		regs.cx--;
	} 
}

static void do_repne_opcodes(const BYTE opcode, const BOOLEAN zero_flag_exit_condition)
{
	while(regs.cx) {
		switch(opcode) {
			case CMPS_BYTE:	do_cmps_byte(); break;
			case CMPS_WORD: do_cmps_word(); break;
			case SCAS_BYTE:	do_scas_byte(); break;
			case SCAS_WORD: do_scas_word(); break;
		}
		regs.cx--;
		
		if(flags.zero == zero_flag_exit_condition)
			return;
	}

#if 0
	} while(--regs.cx && (flags.zero != zero_flag_exit_condition));
#endif
}

static void do_rep()
{
	BYTE opcode;

	opcode = fetch();

	if(check_repne_opcode(opcode)) 
		do_repne_opcodes(opcode, FALSE);
	else 	do_only_rep_opcodes(opcode);
}


static void do_repne()
{
	BYTE opcode;

	opcode = fetch();

	if( !check_repne_opcode(opcode)) {
		fprintf(stderr, "Invalid opcode %x in REPNE\n", opcode);
		do_only_rep_opcodes(opcode);	/* try this -- wierd */
	}

	do_repne_opcodes(opcode, TRUE);
}

/* this is a guess... */
static void load_effective_address()
{
	BYTE secondary_opcode, rm, mod;	
	BYTE secondary_byte;	
	short offset;

	get_secondary_byte();

	*word_reg_table[secondary_opcode >> 3] = compute_effective_address_offset(mod, rm);		
}

/* I don't think we can just add to the current instruction pointer, since we have to
 * fetch the immediate stuff first...
 */
static void do_jump_short_imm(void)
{
	char add = fetch();

	(short) regs.ip += add;  /* signed arithmatic? */
}

static void do_jump_near_imm(void)
{
	short add = fetch_word();

	(short) regs.ip += add; 
}

static void do_jump_far_imm(void)
{
	WORD new_ip;

	new_ip = fetch_word();
	set_selector(CS_SEG, fetch_word());
	regs.ip = new_ip;
}

static void do_hlt()
{
	longjmp(after_halt,  1);
}


static void do_aaa(void)
{
	/* not keeping track of aux carry */
	if((regs.al & 0xf) > 9) {
		regs.al += 6;
		regs.ah++;
		flags.carry = TRUE;
	}
	regs.al &= 0xf;
}

static void do_aad()
{
	if(fetch() != 0xa) {
		fprintf(stderr, "expected 0xa in %s\n", __FUNCTION__);
	}

	regs.al += regs.ah * 0xa;
	regs.ah = 0;
}


static void do_aam(void)
{
	if(fetch() != 0xa) {
		fprintf(stderr, "expected 0xa in %s\n", __FUNCTION__);
	}
 	regs.ah = regs.al/0xa;
	regs.al = regs.al % 0xa;
}

static void do_aas(void)
{
	not_yet_implemented(__FUNCTION__);	
}


static void do_leave()
{

	regs.sp = regs.bp;
	regs.bp = pop();
}

static BYTE *get_immediate_address()
{
	WORD disp = fetch_word();
	BYTE *p = fast_ds;

	switch(seg_override) {
		case  CS_SEG:
			p = fast_cs;
			break;
		case  ES_SEG:
			p = fast_es;
			break;
		case  SS_SEG:
			p = fast_ss;
			break;
	}
	/* default is ds... */
	return p + disp;
}	
 	
static void mov_byte_al_from_xb()
{
	BYTE *address = get_immediate_address();
	
	regs.al = *address;
}

static void mov_word_ax_from_xb()
{
	BYTE *address = get_immediate_address();

	regs.ax = read_pc_word(address);
}

static void mov_byte_al_to_xb()
{
	BYTE *address = get_immediate_address();

	*address = regs.al;
}

static void mov_word_ax_to_xb()
{
	BYTE *address = get_immediate_address();
	write_pc_word(address, regs.ax);
}

/* convert word to doubleword (DX:AX from AX) */
static void do_cwd()
{
	regs.dx = (regs.ax & 0x8000) ? 0xffff : 0;
}

/* this is supposed to be opcode 0x8f -- anything else?? */
static void pop_ea(void)
{
	BYTE secondary_opcode, rm, mod;
        EA effective_address;  
        BYTE secondary_byte;

	get_secondary_byte();

	effective_address = compute_effective_address_word(mod, rm);
	switch(secondary_opcode) {
		case 0:
			write_word(effective_address, pop());
			break;
		default:
			fatal_error();
	}
}

/* do the xlat instruction */
static void do_xlat(void)
{
	BYTE *p;

	/* segment override? */
	p = fast_ds;
	
	p+= regs.bx;
	p += regs.al;
	
	regs.al = *p;
}

static void do_interrupt(BYTE number)
{
	INTERRUPT *pinterrupt = &interrupt_table[number];

	if(number == 0x2f) {
		return;	
	} 
	switch(pinterrupt->type_int) {
		case  PC_INTERRUPT:	
			push(get_flags_directly());
			push(regs.cs);
			push(regs.ip);
			flags.trap = FALSE;
			set_selector(CS_SEG, *(WORD *) &pc_ram[number*4 + 2]);
			regs.ip = *(WORD *) &pc_ram[number*4];
			break;
		case EMULATED_INTERRUPT:
#ifdef TRACE_CALLS
		
			if(number == 0x21 && do_call_trace == TRUE) {
				call_trace("int21:\n");
				print_regs(call_trace_file, &regs);	
				call_trace("\n");
			}
#endif
			(*pinterrupt->int_ptr)(number);
			break;
		default:
			fatal_error();
	} 
}

static void iret()
{
	regs.ip = pop();
	set_selector(CS_SEG, pop());
	set_flags_directly(pop());
}



static inline void do_lds_and_les(SEG seg)
{
	BYTE secondary_opcode, rm, mod;
        EA ea;  
        BYTE secondary_byte;
	WORD segment, offset;


	get_secondary_byte();
	ea = compute_effective_address_byte(mod, rm);
	assert(ea.is_reg == FALSE);

	offset = read_pc_word(ea.ptr);
	segment = read_pc_word(ea.ptr + 2);

	*word_reg_table[secondary_opcode >>3] = offset;
	set_selector(seg, segment);		
}

static inline void do_les(void)
{
	do_lds_and_les(ES_SEG);	
}

static inline void do_lds(void)
{
	do_lds_and_les(DS_SEG);
}
		

static inline void push_imm_word(void)
{
	WORD word;

	word = fetch_word();
	push(word);
}

static inline void push_imm_byte(void)
{
	WORD word;

	word = (WORD ) fetch();
	push(word);
}

WORD last_ip;

emulate_instruction(void)
{
	
	BYTE opcode;

	opcode = fetch();
	save_instruction(opcode);
	switch (opcode){
	    case POP_AX:
		    regs.ax = pop();
		    break;
	    case POP_BX:
		    regs.bx = pop();
		    break;
	    case POP_CX:
		    regs.cx = pop();
		    break;
	    case POP_DX:
		    regs.dx = pop();
		    break;
	    case POP_DI:
		    regs.di = pop();
		    break;
	    case POP_SI:
		    regs.si = pop();
		    break;
	    case POP_BP:
		    regs.bp = pop();
		    break;
	    case POP_SP:
		    regs.sp = pop();
		    break;
	    case POP_DS:
		    set_selector(DS_SEG, pop());
		    break;
	    case POP_ES:
		    set_selector(ES_SEG, pop());
		    break;
	    case POP_SS:
		    set_selector(SS_SEG, pop());
		    break;
	    case PUSH_AX:
		    push(regs.ax);
		    break;
	    case PUSH_BX:
		    push(regs.bx);
			break;
	    case PUSH_CX:
		    push(regs.cx);
		    break;
	    case PUSH_DX:
		    push(regs.dx);
		    break;
	    case PUSH_DI:
		    push(regs.di);
		    break;
	    case PUSH_SI:
		    push(regs.si);
		    break;
	    case PUSH_BP:
		    push(regs.bp);
		    break;
	    case PUSH_SP:
	 	    /* one of the difference of 8086 and 80826 */
#if 1
			/* 8086 strategy */
		    push(regs.sp - 2);
#else
			/* > 286 strategy */
	 	    push(regs.sp);
#endif
		    break;
	    case PUSH_ES:
		    push(regs.es);
		    break;
	    case PUSH_DS:
		    push(regs.ds);
		    break;
	    case PUSH_SS:
		    push(regs.ss);
		    break;
	    case POP_EA:
		    pop_ea();
		    break;
	    case PUSH_CS:
		    push(regs.cs);
		    break;
	    case NOP:
		    do_nop(); 	/* this will be optimized out -- useful for setting breakpoints in gdb */
		    break;
	    case POPF:
		    set_flags_directly(pop());
		    break;
	    case PUSHF:
		    push(get_flags_directly());
		    break;
	    case SAHF:
		    sahf_flags();
		    break;
	    case LAHF:
		    lahf_flags();
		    break;
		    /* INC opcodes */
	    /* increment register */
	    case INC_AX:
		    increment_reg(&regs.ax);
		    break;
	    case INC_BX:
		    increment_reg(&regs.bx);
		    break;
	    case INC_CX:
		    increment_reg(&regs.cx);
		    break;
	    case INC_DX:
		    increment_reg(&regs.dx);
		    break;
	    case INC_SI:
		    increment_reg(&regs.si);
		    break;
	    case INC_DI:
		    increment_reg(&regs.di);
		    break;
	    case INC_BP:
		    increment_reg(&regs.bp);
		    break;
	    case INC_SP:
		    increment_reg(&regs.sp);
		    break;

	    /* decrement register */
	    case DEC_AX:
		    decrement_reg(&regs.ax);
		    break;
	    case DEC_BX:
		    decrement_reg(&regs.bx);
		    break;
	    case DEC_CX:
		    decrement_reg(&regs.cx);
		    break;
	    case DEC_DX:
		    decrement_reg(&regs.dx);
		    break;
	    case DEC_SI:
		    decrement_reg(&regs.si);
		    break;
	    case DEC_DI:
		    decrement_reg(&regs.di);
		    break;
	    case DEC_BP:
		    decrement_reg(&regs.bp);
		    break;
	    case DEC_SP:
		    decrement_reg(&regs.sp);
		    break;


		    /* set/clear flag opcodes */
	    case STD:
		    flags.direction = TRUE;
		    break;
	    case STC:
		    flags.carry = TRUE;
		    break;
	    case STI:
		    flags.interrupt_enable = TRUE;
		    break;
	    case CLI:
		    flags.interrupt_enable = FALSE;
		    break;
	    case CLC:
		    flags.carry = FALSE;
		    break;
	    case CMC:	
		   flags.carry = (flags.carry == TRUE) ? FALSE: TRUE;
		   break;
	    case XLAT:
		   do_xlat();
		   break;
	    case CLD:
		    flags.direction = FALSE;
		    break;
	    case RETURN_FAR:
		    return_far();
		    break;
	    case RETURN_NEAR:
		   return_near();
		   break;
	    case RETURN_NEAR_POP:
		return_near_pop();
		 break;
	    case RETURN_FAR_POP:
		return_far_pop();
		break;

	    /* mov byte/word indirect to/from AL/AX */
	    case MOV_AL_XB:	mov_byte_al_from_xb();	break;
	    case MOV_AX_XB:	mov_word_ax_from_xb();	break;
	    case MOV_XB_AL:	mov_byte_al_to_xb();    break;	
	    case MOV_XB_AX:	mov_word_ax_to_xb();	break;
		  		
	    /* OR opcodes */
  	    case OR_TO_EABYTE:  reg_to_eabyte(or_byte);	break;
	    case OR_TO_EAWORD:	reg_to_eaword(or_word);	break;
	    case OR_TO_BYTEREG:	eabyte_to_reg(or_byte); break;
	    case OR_TO_WORDREG:	eaword_to_reg(or_word); break;
	    case OR_TO_AL:	imm_to_al_byte(or_byte);	    break;
	    case OR_TO_AX:	imm_to_ax_word(or_word);	    break;


	    /* AND opcodes */
            case AND_TO_EABYTE:  reg_to_eabyte(and_byte); break;
            case AND_TO_EAWORD:  reg_to_eaword(and_word); break;
            case AND_TO_BYTEREG: eabyte_to_reg(and_byte); break;
            case AND_TO_WORDREG: eaword_to_reg(and_word); break;
            case AND_TO_AL:      imm_to_al_byte(and_byte);            break;
            case AND_TO_AX:      imm_to_ax_word(and_word);            break;

	    /* ascii adjust */
	    case AAA:		 do_aaa(); break;
	    case AAD:		 do_aad(); break;
	    case AAM:	         do_aam(); break;
	    case AAS:		 do_aas(); break;

            /* SUB opcodes */
            case SUB_TO_EABYTE:  reg_to_eabyte(sub_byte); break;
            case SUB_TO_EAWORD:  reg_to_eaword(sub_word); break;
            case SUB_TO_BYTEREG: eabyte_to_reg(sub_byte); break;
            case SUB_TO_WORDREG: eaword_to_reg(sub_word); break;
            case SUB_TO_AL:      imm_to_al_byte(sub_byte);            break;
            case SUB_TO_AX:      imm_to_ax_word(sub_word);            break;

            /* SBB opcodes */
            case SBB_TO_EABYTE:  reg_to_eabyte(sbb_byte); break;
            case SBB_TO_EAWORD:  reg_to_eaword(sbb_word); break;
            case SBB_TO_BYTEREG: eabyte_to_reg(sbb_byte); break;
            case SBB_TO_WORDREG: eaword_to_reg(sbb_word); break;
            case SBB_TO_AL:      imm_to_al_byte(sbb_byte);            break;
            case SBB_TO_AX:      imm_to_ax_word(sbb_word);            break;

            /* CMP opcodes */
            case CMP_TO_EABYTE:  reg_to_eabyte(cmp_byte); break;
            case CMP_TO_EAWORD:  reg_to_eaword(cmp_word); break;
            case CMP_TO_BYTEREG: eabyte_to_reg(cmp_byte); break;
            case CMP_TO_WORDREG: eaword_to_reg(cmp_word); break;
            case CMP_TO_AL:      imm_to_al_byte(cmp_byte);            break;
            case CMP_TO_AX:      imm_to_ax_word(cmp_word);            break;

	    /* test opcodes */
            case TEST_TO_EABYTE:  reg_to_eabyte(test_byte); break;
            case TEST_TO_EAWORD:  reg_to_eaword(test_word); break;
            case TEST_TO_AL:      imm_to_al_byte(test_byte);            break;
            case TEST_TO_AX:      imm_to_ax_word(test_word);            break;

	    /* XCHG opcodes */
	    case XCHG_BYTE:	reg_to_eabyte(xchg_byte); break;	
	    case XCHG_WORD:	reg_to_eaword(xchg_word); break;
	    case XCHG_BX:	xchg_word(&regs.ax, &regs.bx); break;
	    case XCHG_CX:	xchg_word(&regs.ax, &regs.cx); break;
	    case XCHG_DX:	xchg_word(&regs.ax, &regs.dx); break;
	    case XCHG_BP:	xchg_word(&regs.ax, &regs.bp); break;
	    case XCHG_SI:	xchg_word(&regs.ax, &regs.si); break;
	    case XCHG_DI:	xchg_word(&regs.ax, &regs.di); break;
	    case XCHG_SP:	xchg_word(&regs.ax, &regs.sp); break;
	
	    /* XOR Opcodes */
            case XOR_TO_EABYTE:  reg_to_eabyte(xor_byte); break;
            case XOR_TO_EAWORD:  reg_to_eaword(xor_word); break;
            case XOR_TO_BYTEREG: eabyte_to_reg(xor_byte); break;
            case XOR_TO_WORDREG: eaword_to_reg(xor_word); break;
            case XOR_TO_AL:      imm_to_al_byte(xor_byte);            break;
            case XOR_TO_AX:      imm_to_ax_word(xor_word);            break;

	    /* jump opcodes */
	    case JO:	jump_fetch(get_overflow() == TRUE); break;
	    case JNO:	jump_fetch(get_overflow() == FALSE); break;
	    case JL:	jump_fetch(flags.sign != get_overflow()); break;
            case JNL:	jump_fetch(flags.sign == get_overflow()); break;
	    case JLE:	jump_fetch((flags.sign != get_overflow()) || flags.zero == TRUE); break;
	    case JNLE:	jump_fetch((flags.sign == get_overflow()) && flags.zero == FALSE); break;
	    case JC:	jump_fetch(flags.carry == TRUE); break;
	    case JNC:	jump_fetch(flags.carry == FALSE ); break;
	    case JZ:	jump_fetch(flags.zero == TRUE); break;
	    case JNZ:	jump_fetch(flags.zero == FALSE); break;
	    case JBE:	jump_fetch((flags.carry == TRUE) || (flags.zero == TRUE)); break;
	    case JA:	jump_fetch((flags.carry  == FALSE) && (flags.zero == FALSE)); break;
	    case JS:	jump_fetch(flags.sign == TRUE); break;
	    case JNS:	jump_fetch(flags.sign == FALSE); break;
	    case JP:	jump_fetch(flags.parity == TRUE); break;
	    case JNP:	jump_fetch(flags.parity == FALSE); break;
	    case JCXZ:	jump_fetch(!regs.cx); break;

	    case SEG_OVERRIDE_CS: do_seg_override(CS_SEG); break;
	    case SEG_OVERRIDE_DS: do_seg_override(DS_SEG); break;
	    case SEG_OVERRIDE_ES: do_seg_override(ES_SEG); break;
	    case SEG_OVERRIDE_SS: do_seg_override(SS_SEG); break;

	    case LEA:	load_effective_address(); break;
	    /* move immediate into register */
	    case MOV_AX:	regs.ax = fetch_word(); break;
	    case MOV_BX:	regs.bx = fetch_word(); break;
	    case MOV_CX:	regs.cx = fetch_word(); break;
	    case MOV_DX:	regs.dx = fetch_word(); break;
	    case MOV_BP:	regs.bp = fetch_word(); break;
	    case MOV_SP:	regs.sp = fetch_word(); break;
	    case MOV_SI:	regs.si = fetch_word(); break;
	    case MOV_DI:	regs.di = fetch_word(); break;
	    case MOV_AH:	regs.ah = fetch(); break;
	    case MOV_AL:	regs.al = fetch(); break;
            case MOV_BH:        regs.bh = fetch(); break;
	    case MOV_BL:        regs.bl = fetch(); break;
	    case MOV_CH:	regs.ch = fetch(); break;
	    case MOV_CL:	regs.cl = fetch(); break;
            case MOV_DH:        regs.dh = fetch(); break;
	    case MOV_DL:        regs.dl = fetch(); break;

  	    case MOV_TO_EABYTE:	mov_to_eabyte(); break;
	    case MOV_TO_EAWORD: mov_to_eaword(); break;
	    case MOV_TO_BYTEREG: mov_to_bytereg(); break;
	    case MOV_TO_WORDREG: mov_to_wordreg(); break;
	    case MOV_SEG_TO_EA:   mov_seg_to_ea(); break;
	    case MOV_EA_TO_SEG:	  mov_ea_to_seg(); break;
	    case MOV_IMM_BYTE_TO_EA: mov_imm_byte_to_ea(); break;
	    case MOV_IMM_WORD_TO_EA: mov_imm_word_to_ea(); break;
	    case MOVS_BYTE:	do_movs_byte();  break;
	    case MOVS_WORD:	do_movs_word(); break;
	
	    case IMUL_IMM_BYTE: do_imul_imm_byte(); break;
	    case IMUL_IMM_WORD:  do_imul_imm_word(); break;		    
	    case SCAS_BYTE:	do_scas_byte(); break;
	    case SCAS_WORD:	do_scas_word(); break;
	    
	    case CMPS_BYTE:	do_cmps_byte(); break;
	    case CMPS_WORD:	do_cmps_word(); break;
	    
	    case STOS_BYTE:	do_stos_byte(); break;
	    case STOS_WORD:	do_stos_word(); break;
	   
	    case LODS_BYTE:	do_lods_byte(); break;
	    case LODS_WORD:	do_lods_word(); break;
 
	    /* in/out instructions */
	    case INS_BYTE:	do_ins_byte(); break;
	    case INS_WORD:	do_ins_word(); break;

	    case OUTS_BYTE:	do_outs_byte(); break;
	    case OUTS_WORD:	do_outs_byte(); break;

	    case IN_BYTE_IMM:	in_byte(fetch()); break;
	    case IN_WORD_IMM:	in_word(fetch()); break;
	    case IN_BYTE_DX:	in_byte(regs.dx); break;
	    case IN_WORD_DX:	in_word(regs.dx); break;
	    case OUT_BYTE_IMM:	out_byte(fetch(), regs.al); break;
	    case OUT_WORD_IMM:	out_word(fetch(), regs.ax); break;
	    case OUT_BYTE_DX:	out_byte(regs.dx, regs.al); break;
	    case OUT_WORD_DX:	out_word(regs.dx, regs.ax); break;

	    case REP:		do_rep(); break;
	    case REPNE:		do_repne(); break;

	    case LOOP:		jump_fetch(--regs.cx); break;
	    case LOOPE:		jump_fetch(--regs.cx && ( flags.zero == TRUE)); break;
	    case LOOPNE:	jump_fetch(--regs.cx && ( flags.zero == FALSE)); break;
	
	    case HLT:		do_hlt(); break;
	    case LEAVE:		do_leave(); break;	    
	    case CALL_NEAR_IMM:	do_call_near_imm(); break;
	    case CALL_FAR_IMM:	do_call_far_imm();	    break;

	    /* look at 2nd instruction for actual opcode */
	    /* ROL/ROR/RCL/RCR/SHLSAL/SHR/SAR opcodes */
	    case 0xc0: logical_byte_imm(); break;
	    case 0xc1: logical_word_imm(); break;
	    case 0xd0: logical_byte_once();   break;
	    case 0xd1: logical_word_once();   break;
	    case 0xd2: logical_byte_cl();     break;
	    case 0xd3: logical_word_cl();     break;

	    case 0xf6: do_f6_byte_opcodes();  break;
	    case 0xf7: do_f7_word_opcodes();   break; 
	    case 0xfe: do_fe_byte_opcodes();  break;
	    case 0xff: do_ff_word_opcodes();  break;
	    case 0x80: do_80_byte_opcodes(); break;
	    case 0x81: do_81_word_opcodes(); break;
	    case 0x83: do_83_byte_into_word_opcodes(); break;
	    
	    /* jumps */
	    case JUMP_SHORT_IMM: do_jump_short_imm(); break;
	    case JUMP_NEAR_IMM:	 do_jump_near_imm(); break;
	    case JUMP_FAR_IMM:   do_jump_far_imm(); break;

	    /* adds */
	    case ADD_TO_EA_BYTE:	reg_to_eabyte(add_byte); break;
	    case ADD_TO_EA_WORD:	reg_to_eaword(add_word); break;
	    case ADD_FROM_EA_BYTE:	eabyte_to_reg(add_byte); break;
	    case ADD_FROM_EA_WORD:	eaword_to_reg(add_word); break;
	    case ADD_IMBYTE_TO_AL:	imm_to_al_byte(add_byte); break;
	    case ADD_IMWORD_TO_AX:	imm_to_ax_word(add_word); break;	

	    /* ADC */
	    case ADC_TO_EA_BYTE:	reg_to_eabyte(adc_byte); break;
	    case ADC_TO_EA_WORD:	reg_to_eaword(adc_word); break;
	    case ADC_FROM_EA_BYTE:	eabyte_to_reg(adc_byte); break;
	    case ADC_FROM_EA_WORD:	eaword_to_reg(adc_word); break;
	    case ADC_IMBYTE_TO_AL:	imm_to_al_byte(adc_byte); break;
	    case ADC_IMWORD_TO_AX:	imm_to_ax_word(adc_word); break;	
	   
	    case CBW:	regs.ax = (char) regs.al; break;
	    case CWD:   do_cwd();	break; 
	
	    case INT_IMM: do_interrupt(fetch()); break;
	    case LDS:  do_lds();	break;
	    case LES:  do_les();	break;
	    case IRET:	  iret();		break;
	
	    case PUSH_IMM_BYTE:  push_imm_byte(); break;
	    case PUSH_IMM_WORD:  push_imm_word(); break;

            default:
		    fprintf(stderr, "Can't match opcode %x\n", opcode);	
		    fprintf(stderr, "Ran %d instructions, at %04x:%04x\n",
				num_instructions, regs.cs, regs.ip);
		    fatal_error();

	}
}


static const WORD dummy_segment = 0x0;
static const WORD dummy_offset = 0x2000;

/* this is for dos call 38 */
static void setup_dummy(int number)
{
	BYTE *cp = &pc_ram[(dummy_segment << 4) + dummy_offset];

	*cp++ = INT_IMM;
	*cp++ = (BYTE) number;
	*cp = RETURN_FAR;
}

/* slot in the address (offset, segment) into specified address */
void set_dummy_address(BYTE *p)
{
	write_pc_word(p, dummy_offset);
	write_pc_word(p + 2, dummy_segment);
}

/* this is merely to drap on it in the debugger */
static void dummy_interrupt(void)
{
}	
static void set_interrupt_table()
{	
	enum { VID_BIOS_INT = 0x10, KEYBOARD_BIOS_INT = 0x16 };

	INTERRUPT *pinterrupt;
	
	for(pinterrupt = &interrupt_table; pinterrupt < &interrupt_table[256]; pinterrupt++) {
		pinterrupt->type_int = EMULATED_INTERRUPT;
		pinterrupt->int_ptr =  bad_interrupt; 
	}

	/* DOS interrupt */
	interrupt_table[0x21].type_int = EMULATED_INTERRUPT;
	interrupt_table[0x21].int_ptr = emulate_int21;	
	
	/* int 20 exit */
	interrupt_table[0x20].type_int = EMULATED_INTERRUPT;
	interrupt_table[0x20].int_ptr = process_exit;

	/* dummy interrupt for tracing (we need a call far somewhere */
	interrupt_table[0x80].type_int = EMULATED_INTERRUPT;
	interrupt_table[0x80].int_ptr = dummy_interrupt;
	setup_dummy(0x80);

	interrupt_table[MAGIC_INTERRUPT].type_int = EMULATED_INTERRUPT;
	interrupt_table[MAGIC_INTERRUPT].int_ptr = unix_magic;
	
	interrupt_table[VID_BIOS_INT].type_int = EMULATED_INTERRUPT;
	interrupt_table[VID_BIOS_INT].int_ptr = vid_bios_interrupt;
	
	interrupt_table[KEYBOARD_BIOS_INT].type_int = EMULATED_INTERRUPT;
	interrupt_table[KEYBOARD_BIOS_INT].int_ptr = keyboard_bios_interrupt;
}

/* set up selectors for CS, DS, ES and SS */
void set_up_regs_for_com_file(const int address)
{
	set_selector(CS_SEG, address);
	set_selector(ES_SEG, address);
	set_selector(DS_SEG, address);
	set_selector(SS_SEG, address);
}
 		
void boot_processor(void)
{
	regs = boot_regs;	
	set_selector(CS_SEG, regs.cs);
	set_selector(DS_SEG, regs.ds);
	set_selector(SS_SEG, regs.ss);
	set_selector(ES_SEG, regs.es);
	set_flags_directly(2);
	flags.interrupt_enable = TRUE;
	set_interrupt_table();
#ifdef 0
	boot_msdos();
#endif
}

#if 0
/* bring the file into ram */
load_file()
{
	FILE *fp;
	int c;
	char *p = &pc_ram[boot_regs.cs << 4];
	char *cp;

	fp = fopen(exe_file, "r");
	if(!fp) {
		fprintf(stderr, "can't open %s\n", exe_file);
		return 1;
	}
	
	cp = rindex(exe_file, '.');
	if(cp &&  !strcmp(cp, ".exe")) {
		if(fseek(fp, (long) 512, 0)) {
			perror("fseek failed");
			return;
		}
	}

	while((c = fgetc(fp)) != EOF) 
		*p++ = c;	
	fclose(fp);
	printf("Read %d bytes\n", p - (char *) &pc_ram[boot_regs.cs << 4]);
	return 0;	
}

/* assing a register in the form cc=0xdddd where cc is: cs, ds, es, ss, sp, ip 
 * return TRUE if it works, FALSE otherwise
 */
static BOOLEAN try_to_parse_arg(const char *arg)
{
	WORD_REG *preg = NULL;
	unsigned temp;
	int i;

	if(!strncmp(arg, "cs", 2)) 
		preg = &boot_regs.cs;
	else if(!strncmp(arg, "ds", 2))
		preg = &boot_regs.ds;
	else if(!strncmp(arg, "es", 2))
		preg = &boot_regs.es;
	else if(!strncmp(arg, "ss", 2))
		preg = &boot_regs.ss;
	else if(!strncmp(arg, "sp", 2))
		preg = &boot_regs.sp;

	if(!preg)
		return FALSE;	/* no match */
	
	i = sscanf(arg + 2, "=0x%x", &temp)	;
	if(i != 1)
		return FALSE;
	*preg = temp;
	return TRUE;	
}

/* get registers to boot with from a file called called filename.trc
 * Return FALSE for problems, otherwise TRUE.
 */
static BOOLEAN get_aztec_trc_registers(const char *filename)
{
	char trc_file[80];
	char buffer[100];
	FILE *trace;
	char *cp;
	int i;
	int sp, ss, ds, es, cs;
	int ip;	
	strcpy(trc_file, filename);
	strcat(trc_file, ".trc");	
	trace = fopen(trc_file, "r");
	if(trace == NULL) {
		fprintf(stderr, "Can't open %s\n", trc_file);
		return FALSE;
	}

	/* get sp from first line */
	fgets(buffer, sizeof(buffer), trace);
	cp = index(buffer, 's');
	if(strncmp(cp, "sp=", 3)) {
		fprintf(stderr, "Can't find sp in %s\n", buffer);
		return FALSE;
	}

	i =sscanf(cp+3, "%x", &sp);
	if(i != 1) {
		fprintf(stderr, "Can't find sp in %s\n", buffer);
		return FALSE;
	}
	
	/* get next line */
	fgets(buffer, sizeof(buffer), trace);
	i = sscanf(buffer, "cs=%x  ds=%x  ss=%x  es=%x  ip=%x", 
		&cs, &ds, &ss, &es, &ip);
	if(i != 5) {
		fprintf(stderr, "Can't convert %s\n", buffer);
		return FALSE;
	}
	boot_regs.sp = sp;
	boot_regs.cs = cs;
	boot_regs.ds = ds;
	boot_regs.es = es;
	boot_regs.ss = ss;
	/*boot_regs.ip = ip; */
	return TRUE;
}


static void segmentation_fault()
{
	fprintf(stderr, "caught segmentation fault\n");
	exit();
}

static void ignore_math(void)
{
	puts(__FUNCTION__);
	printf("Instruction %d: cs:ip = %x:%x\n", num_instructions, regs.cs, regs.ip);
}

main(int argc, char **argv) 
{
		signal(SIGSEGV, segmentation_fault);
		signal(SIGFPE,  ignore_math);

                if(setjmp(on_fatal_error)) 
                        abort();
                if(setjmp(after_halt)) 
                      	exit();

		printf("DOS EMULATOR PID = %d\n", getpid());
		
		try_to_read_symbol_table(argv[1]);

		if(boot_msdos(argv[1], argv[2] ) == FALSE) {
			fprintf(stderr, "can't boot msdos");
			exit(1);
		}
		if(rindex(argv[1], '/'))
			exe_file = rindex(argv[1], '/') + 1;
		else	exe_file = argv[1];
		boot_processor();
		while(1) {
			test_breakpoint();
			print_like_aztec();
			emulate_instruction();
		}
} 

#endif

