/* These are opcodes for an msdos emulator
 *
 *   Copyright (C) 1992  Martin Leisner  leisner@eso.mc.xerox.com 
 * 
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License
 *      as published by the Free Software Foundation; either version 2
 *      of the License, or (at your option) any later version.
 * 
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 * 
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 */

#define SECONDARY_MASK	0x38	/* mask to apply to get secondary opcode */

#define AAA		0x37
#define AAD		0xd5     /* 2 byte opcode */

#define AAM		0xd4	/* 2nd byte 0x0a */
#define AAS		0x3f

#define	CBW		0x98
#define CWD		0x99

#define BOUND		0x62

#define CALL_NEAR_IMM	0xe8
#define CALL_FAR_IMM	0x9a

#define JUMP_SHORT_IMM	0xeb
#define JUMP_NEAR_IMM	0xe9
#define JUMP_FAR_IMM	0xea
#define INT_3   0xcc
#define INT_IMM 0xcd
#define INT0    0xce
#define IRET    0xcf

#define MOV_AL_XB	0xa0
#define MOV_AX_XB	0xa1
#define MOV_XB_AL	0xa2
#define MOV_XB_AX	0xa3


#define LEA	0x8d

/* Return instructions */
#define RETURN_FAR		0xcb
#define RETURN_NEAR		0xc3
#define RETURN_NEAR_POP		0xC2
#define RETURN_FAR_POP		0xca

#define PUSH_IMM_BYTE   0x6a
#define PUSH_IMM_WORD	0x68
#define PUSHF		0x9C
#define POPF		0x9D
#define SAHF		0x9e
#define LAHF		0x9f

#define POP_EA		0x8f
#define STC	0xf9
#define STD	0xfd
#define STI	0xfb

#define CLI	0xfa
#define CLC	0xf8
#define CLD	0xfc

/* logical operations use byte bits 3-5 to help figure out the opcode */
#define ROL	(0)
#define ROR	(1 << 3)
#define RCL	(2 << 3)
#define RCR	(3 << 3)
#define SHL_SAL (4 << 3)
#define SHR	(5 << 3)
#define SAR	(7 << 3)
#define NOP		0x90

/* 0x80/0x81 opcodes -- bits 3-5 are used to figure this out */
#define ADD	0
#define OR	(1 << 3)
#define ADC	(2 << 3)
#define SBB	(3 << 3)
#define AND 	(4 << 3)
#define SUB	(5 << 3)
#define XOR 	(6 << 3)
#define CMP	(7 << 3)

/* 0xf6/0xf7 opcodes -- bits 3-5 are used to figure this out */
#define TEST 	0
#define NOT	(2 << 3)
#define NEG	(3 << 3)
#define MUL	(4 << 3)
#define IMUL	(5 << 3)
#define DIV	(6 << 3)
#define IDIV 	(7 << 3)

/* INC (eb)/(ew) | DEC (eb)/word for opcodes 0xfe/0xff */
#define INC_EA	0
#define DEC_EA  (1 << 3)
#define PUSH_EA (6 << 3)
#define CALL_NEAR_EA (2 << 3)
#define CALL_FAR_EA  (3 << 3)
#define JMP_NEAR_EA  (4 << 3)
#define JMP_FAR_EA   (5 << 3)

/* Jump Instructions -- jcond dest where dest is a byte */
#define JO	0x70
#define JNO	0x71
#define JC	0x72
#define JNC	0x73
#define JZ	0x74
#define JNZ	0x75
#define JBE	0x76
#define JA	0x77
#define JS	0x78
#define JNS	0x79
#define JP	0x7a
#define JNP	0x7b
#define JL	0x7c
#define JNL	0x7d
#define JLE	0x7e
#define JNLE	0x7f
#define JCXZ	0xe3

#define XCHG_BYTE	0x86
#define XCHG_WORD	0x87

/* OR opcodes */
#define OR_TO_EABYTE	0x8
#define OR_TO_EAWORD	0x9
#define OR_TO_BYTEREG	0xa
#define OR_TO_WORDREG	0xb
#define OR_TO_AL	0xc
#define OR_TO_AX	0xd

/* ADD opcodes */
#define ADD_TO_EA_BYTE 0x0
#define ADD_TO_EA_WORD 0x1
#define ADD_FROM_EA_BYTE 0x2
#define ADD_FROM_EA_WORD 0x3
#define ADD_IMBYTE_TO_AL	0x4
#define ADD_IMWORD_TO_AX	0x5

/* ADC ocpodes */
#define ADC_TO_EA_BYTE 0x10
#define ADC_TO_EA_WORD 0x11
#define ADC_FROM_EA_BYTE 0x12
#define ADC_FROM_EA_WORD 0x13
#define ADC_IMBYTE_TO_AL	0x14
#define ADC_IMWORD_TO_AX	0x15

/* AND OPCODES */
#define AND_TO_EABYTE    0x20
#define AND_TO_EAWORD    0x21
#define AND_TO_BYTEREG   0x22
#define AND_TO_WORDREG   0x23
#define AND_TO_AL        0x24
#define AND_TO_AX        0x25

/* SUB OPCODES */
#define SUB_TO_EABYTE    0x28
#define SUB_TO_EAWORD    0x29
#define SUB_TO_BYTEREG   0x2a
#define SUB_TO_WORDREG   0x2b
#define SUB_TO_AL        0x2c
#define SUB_TO_AX        0x2d

/* SBB OPCODES */
#define SBB_TO_EABYTE    0x18
#define SBB_TO_EAWORD    0x19
#define SBB_TO_BYTEREG   0x1a
#define SBB_TO_WORDREG   0x1b
#define SBB_TO_AL        0x1c
#define SBB_TO_AX        0x1d

/* CMP OPCODES */
#define CMP_TO_EABYTE    0x38
#define CMP_TO_EAWORD    0x39
#define CMP_TO_BYTEREG   0x3a	
#define CMP_TO_WORDREG   0x3b
#define CMP_TO_AL        0x3c
#define CMP_TO_AX        0x3d


/* TEST opcodes */
#define TEST_TO_EABYTE    0x84
#define TEST_TO_EAWORD    0x85
#define TEST_TO_AL        0xa8
#define TEST_TO_AX        0xa9 

/* XOR opcodes */
#define XOR_TO_EABYTE    0x30
#define XOR_TO_EAWORD    0x31
#define XOR_TO_BYTEREG   0x32
#define XOR_TO_WORDREG   0x33
#define XOR_TO_AL        0x34
#define XOR_TO_AX        0x35

/* MOV opcodes */
#define MOV_TO_EABYTE	0x88
#define MOV_TO_EAWORD	0x89
#define MOV_TO_BYTEREG	0x8a
#define MOV_TO_WORDREG	0x8b
#define MOV_SEG_TO_EA	0x8c
#define MOV_EA_TO_SEG	0x8e
#define MOV_IMM_BYTE_TO_EA 0xc6
#define MOV_IMM_WORD_TO_EA 0xc7
#define MOVS_BYTE 0xa4
#define MOVS_WORD	0xa5

#define SCAS_BYTE	0xae
#define SCAS_WORD	0xaf

#define STOS_BYTE	0xaa
#define STOS_WORD	0xab

#define LODS_BYTE 	0xac
#define LODS_WORD	0xad
#define IN_BYTE_IMM	0xe4
#define IN_WORD_IMM	0xe5
#define IN_BYTE_DX	0xec
#define IN_WORD_DX	0xed

#define INS_BYTE 0x6c
#define INS_WORD	0x6d

#define OUT_BYTE_IMM	0xe6
#define OUT_WORD_IMM	0xe7
#define OUT_BYTE_DX	0xee
#define OUT_WORD_DX	0xef

#define OUTS_BYTE	0x6e
#define OUTS_WORD	0x6f

#define CMPS_BYTE 0xa6
#define CMPS_WORD 0xa7

#define HLT	0xf4
#define REP	0xf3
#define REPNE   0xf2
#define LOOP	0xe2
#define LOOPE	0xe1
#define LOOPNE	0xe0

#define IMUL_IMM_BYTE 0x6b
#define IMUL_IMM_WORD  0x69

#define LEAVE 0xc9
#define ENTER  0xc8

#define LDS	0xc5
#define LES	0xc4

#define CMC	0xf5   

#define XLAT	0xd7
