
/* An Interrupt 21 emulator for Ms/Dos
 * 
 *
 *   Copyright (C) 1992  Martin Leisner  leisner@eso.mc.xerox.com 
 * 
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License
 *      as published by the Free Software Foundation; either version 2
 *      of the License, or (at your option) any later version.
 * 
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 * 
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
/*
 * $Log: dosboot.c,v $
 * Revision 1.2  1994/07/22  18:10:39  leisner
 * *** empty log message ***
 *
 * Revision 1.1  1994/07/12  22:31:57  leisner
 * Initial revision
 *
 * Revision 1.1  1993/07/11  21:04:50  leisner
 * Initial revision
 *
 */

#include <stdio.h>
#include <assert.h>
#include <stdarg.h>

#include "misc.h"
#include "8086regs.h"
#include "memory.h"
#include "task.h"
#include "doserrno.h"
#include "dosprotos.h"

static char id[] = "$Id: dosboot.c,v 1.2 1994/07/22 18:10:39 leisner Exp leisner $";
static char compiled_at[] = COMPILED_AT;

static MEM_BLOCK *get_first_environment_from_file(const char *name)
{
	char **environ_strings;    /* malloced for each line */
	int num_strings = 0;
	char line_buf[1000];	/* for each line, knowing \\n is a
				 * continued line 
				 */
	FILE *stream;
	char *cp;
	MEM_BLOCK *pmem;
	int i;
	int length;
	char *string_space;

	stream = fopen(name, "r");
	if(stream == NULL) {
		fprintf(stderr, "can't open %s for environment\n",
			name);
		exit(1);
	}
	environ_strings = malloc(sizeof (char **));	/* no space yet */

	/* get each string, malloc space for it, and put on the end of the array of points to
	 * environment lines
	 */
	while(fgets(line_buf, sizeof line_buf, stream) != NULL) {
		assert(strlen(line_buf) != sizeof line_buf);
		
		for(cp = line_buf; *cp && cp < line_buf + sizeof line_buf; cp++)
			if(*cp == '\n') {
				*cp = '\0';
				break;
			}
		assert(*cp == '\0'); 
		environ_strings = realloc(environ_strings, sizeof(char **) + 1);
		environ_strings[num_strings] = malloc(strlen(line_buf) + 1);
		strcpy(environ_strings[num_strings], line_buf);
		num_strings++;

	}

	assert(num_strings != 0);

	for(i = 0, length = 1; i < num_strings; i++)  { 
		length += strlen(environ_strings[i]) + 1;
	}

	string_space = malloc(length);

	for(cp = string_space, i = 0; i < num_strings;  i++) {
		strcpy(cp, environ_strings[i]);
		while(*cp++);
			;
	}		

	/* end of array has double null */
	*cp = '\0';
	
	pmem = make_environment(string_space);
	fclose(stream);
	free(string_space);
	for(i = 0; i < num_strings - 1; i++)
		free(environ_strings[i]);
	free(environ_strings);	
	return pmem;
}

/* get the first environment.
 * If flag is set for use_disk_get_environment, get the file, otherwise
 * return NULL
 */	
static MEM_BLOCK *get_first_environment(void) 
{

	static char *first_environment =  "PATH=/bin\0\0" ;

	if(dos_environment) {
		/* have an environment file */
		return get_first_environment_from_file(dos_environment);
	} else return make_environment(first_environment);
}

char exec_prog_name[80];

/* boot msdos -- with a com file */
BOOLEAN boot_msdos(char *first_task_name, const char *command_line)
{
	MEM_BLOCK *penvironment;

	boot_memory();
	boot_filesystems();
	strcpy(exec_prog_name, "drcom.com");

	penvironment = get_first_environment();

	if(load_first_task(first_task_name,  penvironment, command_line) == FALSE) {
		fprintf(stderr, "Can't load first task %s\n", first_task_name);
		exit(1);
	}


	initialize_handles_for_first_task(current_task);
	
	/* get current path */
	current_default_drive = 'D' - 'A';

	dos_errno = NO_ERROR;
}

