/* An Interrupt 21 emulator for Ms/Dos
 * 
 *
 *   Copyright (C) 1992  Martin Leisner  leisner@eso.mc.xerox.com 
 * 
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License
 *      as published by the Free Software Foundation; either version 2
 *      of the License, or (at your option) any later version.
 * 
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 * 
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
/*
 * $Log: dosemu.c,v $
 * Revision 1.1  1993/07/11  21:04:50  leisner
 * Initial revision
 *
 */

#include <stdio.h>
#include <assert.h>
#include <stdarg.h>

#include "misc.h"
#include "8086regs.h"
#include "memory.h"
#include "task.h"
#include "doserrno.h"
#include "dosprotos.h"

static char *id = "$Id: dosemu.c,v 1.1 1993/07/11 21:04:50 leisner Exp leisner $";
static char compiled_at[] = COMPILED_AT;

/* interrupt 21 functions -- from ah */
#define TERMINATE_PROGRAM  0x0
#define READ_STDIN_WITH_ECHO 0x1 
#define WRITE_STDOUT 0x2
#define DIRECT_CON_IO 0x06
#define BUFFERED_INPUT   0xa
#define  STDOUT_WRITE_STRING 0x9
#define GET_STDIN_STATUS 0xb
#define  SELECT_DEFAULT_DRIVE 0xe
#define FCB_CLOSE	0x10
#define FCB_FINDFIRST 0x11
#define FCB_FINDNEXT  0x12
#define FCB_DELETE	0x13
#define FCB_SEQ_READ	0x14
#define FCB_SEQ_WRITE	0x15
#define GET_CURRENT_DEFAULT_DRIVE 0x19
#define SET_DTA_ADDRESS 0x1a
#define SET_INTERRUPT_VECTOR 0x25
#define FCB_PARSE_FILENAME 0x29
#define GET_SYSTEM_DATE	0x2a
#define SET_SYSTEM_DATE 0x2b
#define GET_SYSTEM_TIME 0x2c
#define GET_DTA_ADDRESS	0x2f
#define GET_DOS_VERSION 0x30
#define DOS_EXTENDED_BREAK_CHECKING 0x33 
#define GET_INTERRUPT_VECTOR 0x35
#define GET_FREE_DISK_SPACE  0x36
#define SWITCHAR	0x37
#define MKDIR		0x39
#define DOS_COUNTRY_INFO 0x38
#define RMDIR		0x3a
#define CHDIR		0x3b
#define CREATE		0x3c
#define OPEN 		0x3d	
#define CLOSE		0x3e
#define READ		0x3f
#define WRITE		0x40
#define UNLINK		0x41
#define LSEEK		0x42
#define CHMOD		0x43
#define IOCTL	0x44
#define DUP	0x45
#define DUP2	0x46
#define GET_CUR_DIRECTORY  0x47
#define ALLOCATE_BLOCK 0x48
#define FREE_BLOCK     0x49
#define RESIZE_BLOCK 0x4a
#define LOAD	0x4B
#define EXIT	0x4C
#define GET_RETURN_CODE 0x4d
#define FIND_FIRST	0x4e
#define FIND_NEXT	0x4f
#define SET_CURRENT_PSP 0x50
#define GET_CURRENT_PSP_UNDOC 0x51
#define CREATE_CHILD_PSP	0x55
#define GET_CURRENT_PSP 0x62
#define GET_FILE_DATE_AND_TIME 0x57
#define ALLOC_STRATEGY  0x58
#define GET_EXTENDED_ERROR_INFO 0x59
#define INTERNAL_SERVER_FUNC 0x5d
#define CANONICALIZE_FILENAME 0x60
#define GET_INPUT_MODE	0x63
#define GET_EXTENDED_COUNTRY_INFO 0x65

#define SHOW_DOS(x)		case x:	 fprintf(where, "Function %-3x: %-20s", x, #x); break;

void dos_messages( char *fmt, ... )
{
	static FILE *dos_messages = NULL;
	va_list args;

	 if(!dos_messages) {
		dos_messages = fopen("debug.trace", "w");
		if(!dos_messages) {
			perror("Can't open debug.trace");
			exit(1);
		}
	}
 
        va_start(args, fmt);
        vfprintf(dos_messages, fmt, args);
	fflush(dos_messages);
                
        va_end(args);

}

static void show_dos_call(FILE *where, const int call)
{
	switch(call) {
		SHOW_DOS(FCB_PARSE_FILENAME);
		SHOW_DOS(INTERNAL_SERVER_FUNC);
		SHOW_DOS(GET_CURRENT_DEFAULT_DRIVE);
		SHOW_DOS(GET_CUR_DIRECTORY);
		SHOW_DOS(RESIZE_BLOCK);
		SHOW_DOS(FIND_FIRST);
		SHOW_DOS(GET_CURRENT_PSP);
		SHOW_DOS(GET_CURRENT_PSP_UNDOC);
		SHOW_DOS(GET_RETURN_CODE);
		SHOW_DOS(SET_DTA_ADDRESS);
		SHOW_DOS(SET_INTERRUPT_VECTOR);
		SHOW_DOS(MKDIR);
		SHOW_DOS(GET_INPUT_MODE);
		SHOW_DOS(GET_EXTENDED_ERROR_INFO);
		SHOW_DOS(CHMOD);
		SHOW_DOS(GET_DTA_ADDRESS);
		SHOW_DOS(DUP);
		SHOW_DOS(DUP2);
		SHOW_DOS(CHDIR);
		SHOW_DOS(TERMINATE_PROGRAM);
		SHOW_DOS(READ_STDIN_WITH_ECHO);
		SHOW_DOS(WRITE_STDOUT);
		SHOW_DOS(DIRECT_CON_IO);
		SHOW_DOS(BUFFERED_INPUT);
		SHOW_DOS(STDOUT_WRITE_STRING);
		SHOW_DOS(GET_STDIN_STATUS);
		SHOW_DOS(SELECT_DEFAULT_DRIVE);
		SHOW_DOS(FCB_CLOSE);
		SHOW_DOS(FCB_FINDFIRST);
		SHOW_DOS(FCB_FINDNEXT);
		SHOW_DOS(FCB_DELETE);
		SHOW_DOS(READ);
		SHOW_DOS(WRITE);
		SHOW_DOS(UNLINK);	
		SHOW_DOS(LSEEK);
		SHOW_DOS(IOCTL);
		SHOW_DOS(ALLOCATE_BLOCK);
		SHOW_DOS(FREE_BLOCK);
		SHOW_DOS(EXIT);
		SHOW_DOS(LOAD); 
		SHOW_DOS(ALLOC_STRATEGY);
		SHOW_DOS(SET_CURRENT_PSP);
		SHOW_DOS(CANONICALIZE_FILENAME);
		SHOW_DOS(GET_EXTENDED_COUNTRY_INFO);
		SHOW_DOS(GET_SYSTEM_DATE);
		SHOW_DOS(SET_SYSTEM_DATE);
		SHOW_DOS(GET_SYSTEM_TIME);
		SHOW_DOS(GET_DOS_VERSION);
		SHOW_DOS(GET_INTERRUPT_VECTOR);
		SHOW_DOS(GET_FREE_DISK_SPACE);
		SHOW_DOS(SWITCHAR);
		SHOW_DOS(DOS_COUNTRY_INFO);
		SHOW_DOS(RMDIR);
		SHOW_DOS(CREATE);
		SHOW_DOS(OPEN);
		SHOW_DOS(CLOSE);
		SHOW_DOS(GET_FILE_DATE_AND_TIME);
		SHOW_DOS(CREATE_CHILD_PSP);
		SHOW_DOS(FIND_NEXT);
		default:
			fprintf(where, "unknown dos code = %x ", call);
	}

	fprintf(where, "Instruction # %d -- %x:%x\n", num_instructions, regs.cs, regs.ip -2);
}


void trace_dos_call(void)
{

	static FILE *dos_trace = 0;
	if(!dos_trace) {
		dos_trace = fopen("dos.trace", "w");
		if(!dos_trace) {
			fprintf(stderr, "Can't open dos.trace\n");
			perror("");
			exit(1);
		}
	}

	show_dos_call(dos_trace, regs.ah);
	fflush(dos_trace);
}

/* p points to a far pointer (offset:segment) in pc ram.
 * convert it into another pointer into pc ram
 */
BYTE *convert_far_pointer(BYTE *p)
{
	/* use ints instead of words...it may work better? */
	int segment;
	int  offset;

	offset = read_pc_word(p);
	segment = read_pc_word(p + 2);
	return &pc_ram[(segment << 4) + offset];
}

/*
 * INT 21 - DOS 2+ - "SWITCHAR" - GET SWITCH CHARACTER
 *         AX = 3700h
 * Return: AL = status
 *             00h successful
 *                 DL = current switch character
 *             FFh unsupported subfunction
 * Desc:   Determine the character which is used to introduce command switches.
 *           This setting is ignored by DOS commands in version 4.0 and higher,
 *           but is honored by many third-party programs.
 * Notes:  documented in some OEM versions of some releases of DOS
 *         supported by OS/2 compatibility box
 *         always returns DL=2Fh for DOS 5.0
 * SeeAlso: AX=3701h
 * -------------------------------------------------------------------------------
 * INT 21 - DOS 2+ - "SWITCHAR" - SET SWITCH CHARACTER
 *         AX = 3701h
 *         DL = new switch character
 * ^LReturn: AL = status
 *             00h successful
 *             FFh unsupported subfunction
 * Notes:  documented in some OEM versions of some releases of DOS
 *         supported by OS/2 compatibility box
 *         ignored by DOS 5.0
 * SeeAlso: AX=3700h
 * -------------------------------------------------------------------------------
 * INT 21 - DOS 2.x and 3.3+ only - "AVAILDEV" - SPECIFY \DEV\ PREFIX USE
 *         AH = 37h
 *         AL = subfunction
 *             02h get availdev flag
 *                 Return: DL = 00h \DEV\ must precede character device names
 *             03h set availdev flag
 *                 DL = 00h        \DEV\ is mandatory
 *                    = nonzero    \DEV\ is optional
 * Return: AL = status
 *             00h successful
 *             FFh unsupported subfunction
 * Notes:  all versions of DOS from 2.00 allow \DEV\ to be prepended to device
 *           names without generating an error even if the directory \DEV does
 *           not actually exist (other paths generate an error if they do not
 *           exist).
 *         although DOS 3.3+ accepts these calls, they have no effect, and
 *           AL=02h always returns DL=FFh
*/
static void do_switchar(void)
{
	printf("setting switchar, subfunction = %d\n",
			regs.al);

	switch(regs.al) {
		case 0:
			regs.dl = '/';
			regs.al = 0;
			break;
		default:
	}		
}

/*
 * INT 21 - DOS 2+ - GET DOS VERSION
 *         AH = 30h
 * Return: AL = major version number (00h if DOS 1.x)
 *             AX = error code (see AH=59h)
 *         CF clear if successful
 * Note:   not supported by OS/2
 * SeeAlso: INT 2F/AX=1403h
 * AH = minor version number
 * 
 */
static void get_dos_version(void)
{
#define DOS4	4
#define DOS3	0x1e03


	static BOOLEAN first_time = FALSE;
	static WORD version = DOS3;
	if(first_time == FALSE) {
		char *cp;

		first_time = TRUE;
		if(cp = getenv("DOSVERSION")) {
			if(strcmp(cp, "DOS4") == 0) {
				version = DOS4;
				printf("Defaulting to DOS4\n");
			} else  printf("Unknown version in envionment %s\n", cp);
		}
	}
	
	regs.ax = version;		
}


/* 
 *  INT 21 - DOS 3+ - GET OR SET MEMORY ALLOCATION STRATEGY
 *         AH = 58h
 *         AL = subfunction
 *             00h get allocation strategy
 *                 Return: AX = current strategy
 *                             00h low memory first fit
 *                             01h low memory best fit
 *                             02h low memory last fit
 *                          ---DOS 5.0---
 *                             40h high memory first fit
 *                             41h high memory best fit
 *                             42h high memory last fit
 *                             80h first fit, try high then low memory
 *                             81h best fit, try high then low memory
 *                             82h last fit, try high then low memory
 *             01h set allocation strategy
 *                 BL = new allocation strategy (see above)
 *                 BH = 00h (DOS 5.0)
 * Return: CF clear if successful
 *         CF set on error
 *             AX = error code (01h) (see AH=59h)
 * Notes:  the Set subfunction accepts any value in BL for DOS 3.x and 4.x;
 *           2 or greater means last fit
 *         the Get subfunction returns the last value set
 *         a program which changes the allocation strategy should restore it
 *           before terminating
 *         Toshiba MSDOS 2.11 supports subfunctions 00h and 01h
 *         DR-DOS 3.41 reportedly reverses subfunctions 00h and 01h
 */
static void do_mem_alloc_strategy()
{
	if(regs.al == 0) {
		/* get strategy */
		regs.ax = (WORD) mem_alloc_strategy;
		flags.carry = FALSE;	/* must be successful */
	} 
#if 0
else (regs.al == 1) {
		if(regs.bl <= 0x2) {
			mem_alloc_strategy = (MEM_ALLOC_STRATEGY) regs.bl;
			flags.carry = FALSE;
		} else flags.carry = TRUE;	
	} else flags.carry = TRUE;	
#endif
}

/* 
 *
INT 21 - DOS 2+ - GET INTERRUPT VECTOR
        AH = 35h
        AL = interrupt number
Return: ES:BX -> current interrupt handler
 */
void dos_get_interrupt_vector(void)
{
	fprintf(stderr, "getting vector %d\n", regs.al);
}



void bad_interrupt(const int interrupt_number)
{
	fprintf(stderr, "bad interrupt #%xH\n", interrupt_number);
	print_saved_instructions();
#if 0
	abort();
#endif
}

void set_pc_interrupt(int number, WORD segment, WORD offset)
{

	if(number  == 0x23) {
		interrupt_table[number].type_int = PC_INTERRUPT;
		write_pc_word(&pc_ram[number*4], offset);
		write_pc_word(&pc_ram[number * 4 + 2], offset);
	}
}

void set_bad_interrupt(number)
{
	interrupt_table[number].type_int = EMULATED_INTERRUPT;
	interrupt_table[number].int_ptr = bad_interrupt;
}

/* 
 * 
 INT 21 - DOS 1+ - SET INTERRUPT VECTOR
        AH = 25h
        AL = interrupt number
        DS:DX -> new interrupt handler
 */
void dos_set_interrupt_vector(void)
{
	fprintf(stderr, "setting vector %x to %x:%x\n",
		regs.al, regs.ds, regs.dx);

	current_task->interrupt_23 = TRUE;
	current_task->int23_seg = regs.ds;
	current_task->int23_off = regs.dx; 
	set_pc_interrupt(regs.al, regs.ds, regs.dx);
}


/* write the character c to the handle for the current process */
static void write_dos_char(char handle, char c)
{
}

static void direct_con_io()
{
	if(regs.dl) {
		/* input */
	} else { 
		write_dos_char(1, regs.dh);
		regs.al = regs.dh;
	}
}



/*
INT 21 - DOS 2+ - "RMDIR" - REMOVE SUBDIRECTORY
        AH = 3Ah
        DS:DX -> ASCIZ pathname of directory to be removed
Return: CF clear if successful
            AX destroyed
        CF set on error
            AX = error code (03h,05h,06h,10h) (see AH=59h)
*/
static void do_rmdir()
{
	char *name;

	name = convert_dos_asciiz_to_unix_name(fast_ds + regs.dx);

	if(!name)  {
		flags.carry = TRUE;
		return;
	}	

	/* check to make sure we aren't in this directory? */
	if(rmdir(name)) {
		/* error */
		flags.carry = TRUE;
	} else {
		flags.carry = FALSE;
	}	
	
}
/* 
 INT 21 - DOS 2+ - "MKDIR" - CREATE SUBDIRECTORY
        AH = 39h
        DS:DX -> ASCIZ pathname
Return: CF clear if successful
            AX destroyed
        CF set on error
            AX = error code (03h,05h) (see AH=59h)
Notes:  all directories in the given path except the last must exist
        fails if the parent directory is the root and is full
        DOS 2.x-3.3 allow the creation of a directory sufficiently deep that
          it is not possible to make that directory the current directory
          because the path would exceed 64 characters
        under the FlashTek X-32 DOS extender, the pointer is in DS:EDX
SeeAlso: AH=3Ah,AH=3Bh,AH=E2h/SF=0Ah,INT 2F/AX=1103h
*/

static void do_mkdir()
{

	char *name;

	name = convert_dos_asciiz_to_unix_name(fast_ds + regs.dx);

	if(!name)  {
		flags.carry = TRUE;
		return;
	}	

	/* check to make sure we aren't in this directory? */
	if(mkdir(name, 0775)) {
		/* error */
		flags.carry = TRUE;
	} else {
		flags.carry = FALSE;
	}	
	
}



void emulate_int21()
{
	trace_dos_call();	
	switch(regs.ah) {
		case BUFFERED_INPUT:
			dos_buffered_input();
			break;
		case WRITE_STDOUT:
			write_character_to_stdout();
			break;
		case DIRECT_CON_IO:	
			direct_con_io(); break;
		case GET_STDIN_STATUS:
			get_stdin_status();
			break;
	        case SET_SYSTEM_DATE:
			set_dos_date();
			break;
		case GET_SYSTEM_DATE:
			get_dos_date();
			break;
		case GET_SYSTEM_TIME:
			get_dos_time();
			break;
		case SELECT_DEFAULT_DRIVE:
			dos_select_default_drive();
			break;
		case DOS_EXTENDED_BREAK_CHECKING:
			dos_extended_break_checking();
			break;
		case FCB_CLOSE:
			dos_fcb_close();
			break;
		case FCB_FINDFIRST:
			dos_fcb_findfirst();
			break;
		case FCB_FINDNEXT:
			dos_fcb_findnext();
			break;
		case GET_FILE_DATE_AND_TIME:
			dos_get_file_date_and_time();
			break;
	        case LOAD:
			dos_exec();
			break;
		case CHDIR:	
			dos_chdir();
			break;
		case CREATE:
			do_create();
			break;
		case MKDIR:	
			do_mkdir();
			break;
		case RMDIR:
			do_rmdir();
			break;
		case OPEN:
			do_open();
			break;
		case CLOSE:
			do_close();
			break;
		case READ:
			do_read();
			break;
		case SET_INTERRUPT_VECTOR:
			dos_set_interrupt_vector();
			break;
		case GET_INTERRUPT_VECTOR:
			dos_get_interrupt_vector();
			break;
		case WRITE:
			do_write();
			break;
		case UNLINK:
			dos_unlink_asciiz();
			break;
		case GET_EXTENDED_ERROR_INFO:
			get_extended_error_info();
			break;
		case IOCTL:
			do_ioctl();
			break;
	 	case STDOUT_WRITE_STRING:
			standard_write_string();
			break;
		case LSEEK:
			do_lseek();
			break;
		case FCB_PARSE_FILENAME:
			fcb_parse_filename();
			break;
		case GET_FREE_DISK_SPACE:
			get_dos_free_disk_space();
			break;
		case GET_DOS_VERSION:
			get_dos_version();
			break;
		case RESIZE_BLOCK:
			do_resize_block();
			break;
		case ALLOC_STRATEGY:
			do_mem_alloc_strategy();
			break;
		case SWITCHAR:
			do_switchar();
			break;
		case GET_CURRENT_DEFAULT_DRIVE:		
			get_current_default_drive();
			break;
		case FREE_BLOCK:
			dos_free_block();
			break;
		case ALLOCATE_BLOCK:
			dos_get_block();
			break;
		case EXIT:
			process_exit(); break;
		case SET_DTA_ADDRESS:
			dos_set_dta_address(); break;
		case GET_DTA_ADDRESS:
			dos_get_dta_address(); break;
		case FIND_FIRST:
			dos_find_first_file(); break;
		case FIND_NEXT:
			dos_find_next_file();
			break;
		case DUP:		
			dos_dup();
			break;
		case DUP2:
			dos_dup2();
			break;
		case CHMOD:
			dos_chmod();
			break;
		case GET_CUR_DIRECTORY:
			dos_getcwd();
			break;
		case GET_INPUT_MODE:	
			/* not supported */
			regs.al = 0xff;
			break;
		case GET_RETURN_CODE:
			get_process_return_code();
			break;
		case GET_CURRENT_PSP_UNDOC:
		case GET_CURRENT_PSP:
			get_current_psp();
			break;
		case INTERNAL_SERVER_FUNC:
			dos_internal_server_function();
			break;
		case SET_CURRENT_PSP:
			set_current_psp();
			break;
		case CANONICALIZE_FILENAME:
			dos_canonicalize_filename();
			break;
		case DOS_COUNTRY_INFO:
			dos_country_info();
			break;	
		case GET_EXTENDED_COUNTRY_INFO:
			dos_extended_country_info();
			break;
		case CREATE_CHILD_PSP:
			create_child_psp();
			break;
		default:
			fprintf(stderr, "INT21 undefined, %x\n", regs.ah);
			break;
	}
}
