/*
 *   Copyright (C) 1992  Martin Leisner  leisner@eso.mc.xerox.com 
 * 
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License
 *      as published by the Free Software Foundation; either version 2
 *      of the License, or (at your option) any later version.
 * 
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 * 
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <stdio.h>


#include "misc.h"
#include "8086regs.h"
#include "task.h"

static char compiled_at[] = COMPILED_AT;


/* 
 *----------212B-------------------------------
 *INT 21 - DOS 1+ - SET SYSTEM DATE
 *        AH = 2Bh
 *        CX = year (1980-2099)
 *        DH = month
 *        DL = day
 *Return: AL = status
 *            00h successful
 *            FFh invalid date, system date unchanged
 *Note:   DOS 3.3+ also sets CMOS clock
 *SeeAlso: AH=2Ah,AH=2Dh,INT 1A/AH=05h
 */
void set_dos_date(void)
{
	printf("Trying to set dos date\n");
	printf("regs.cx = %d, regs.dh = %d, regs.dl = %d\n",
			regs.cx, regs.dh, regs.dl);
	
	regs.al = 0xff;
}

/*
 * INT 21 - DOS 3.3+ - GET EXTENDED COUNTRY INFORMATION
 *         AH = 65h
 *         AL = info ID
 *             01h get general internationalization info
 *             02h get pointer to uppercase table
 *             04h get pointer to filename uppercase table
 *             05h get pointer to filename terminator table
 *             06h get pointer to collating sequence table
 *             07h (DOS 4+) get pointer to Double-Byte Character Set table
 *         BX = code page (-1=global code page)
 *         DX = country ID (-1=current country)
 *         ES:DI -> country information buffer (see below)
 *         CX = size of buffer (>= 5)
 * Return: CF set on error
 *             AX = error code (see AH=59h)
 *         CF clear if succesful
 *             CX = size of country information returned
 *             ES:DI -> country information
 * Notes:  AL=05h appears to return same info for all countries and codepages; it
 *           has been documented for DOS 5.0, but was undocumented in ealier
 *           versions
 *         NLSFUNC must be installed to get info for countries other than the
 *           default
 *         subfunctions 02h and 04h are identical under OS/2
 * SeeAlso: AH=38h,INT 2F/AX=1401h,INT 2F/AX=1402h
 * 
 * Format of country information:
 * Offset  Size    Description
 *  00h    BYTE    info ID
 * ---if info ID = 01h---
 *  01h    WORD    size
 *  03h    WORD    country ID
 *  05h    WORD    code page
 *  07h 34 BYTEs   country-dependent info (see AH=38h)
 * ---if info ID = 02h---
 *  01h    DWORD   pointer to uppercase table (see below)
 * ---if info ID = 04h---
 *  01h    DWORD   pointer to filename uppercase table (see below)
 * ---if info ID = 05h---
 *  01h    DWORD   pointer to filename character table (see below)
 * ---if info ID = 06h---
 *  01h    DWORD   pointer to collating table (see below)
 * ---if info ID = 07h (DOS 4+)---
 *  01h    DWORD   pointer to DBCS lead byte table (see below)
 * 
 * Format of uppercase table:
 * Offset  Size    Description
 *  00h    WORD    table size
 *  02h 128 BYTEs  uppercase equivalents (if any) of chars 80h to FFh
 * 
 * Format of collating table:
 * Offset  Size    Description
 *  00h    WORD    table size
 *  02h 256 BYTEs  values used to sort characters 00h to FFh
 * 
 * Format of filename terminator table:
 * Offset  Size    Description
 *  00h    WORD    table size (not counting this word)
 *  02h    BYTE    ??? (01h for MSDOS 3.30-5.00)
 *  03h    BYTE    lowest permissible character value for filename
 *  04h    BYTE    highest permissible character value for filename
 *  05h    BYTE    ??? (00h for MSDOS 3.30-5.00)
 *  06h    BYTE    first excluded character in range \ all characters in this
 *  07h    BYTE    last excluded character in range  / range are illegal
 *  08h    BYTE    ??? (02h for MSDOS 3.30-5.00)
 *  09h    BYTE    number of illegal (terminator) characters
 *  0Ah  N BYTES   characters which terminate a filename:  ."/\[]:|<>+=;,
 * Note:   partially documented for DOS 5.0, but undocumented for earlier versions
 * 
 * Format of filename uppercase table:
 * Offset  Size    Description
 *  00h    WORD    table size
 *  02h 128 BYTEs  uppercase equivalents (if any) of chars 80h to FFh
 * 
 * Format of DBCS lead byte table:
 * Offset  Size    Description
 *  00h    WORD    length
 *  02h 2N BYTEs   start/end for N lead byte ranges
 *         WORD    0000h   (end of table)
 */
void dos_extended_country_info(void)
{
	enum { GET_GENERAL = 01,
	       GET_UPPERCASE = 2,
	       GET_FILENAME_UPPERCASE = 4,
	       GET_FILENAME_TERMINATOR = 5,
		GET_DOUBLE_BYTE_TABLE = 7 /* only for dos 4.0 */
	};

	switch(regs.al) {
		case GET_DOUBLE_BYTE_TABLE:
			/* I don't understand this yet */
		       write_pc_word(fast_es + regs.di, 0);
			regs.cx = 0;
			break;	
		case GET_GENERAL:
		case GET_UPPERCASE:
		case GET_FILENAME_UPPERCASE:
		case GET_FILENAME_TERMINATOR:	
		default:
			printf(__FUNCTION__ " with code %d\n", regs.al);
			flags.carry = FALSE;
			return;
	}
	flags.carry = TRUE;
}

/* INT 21 - DOS 3+ - SET COUNTRY CODE
 * INT 21 - DOS 2+ - GET COUNTRY-SPECIFIC INFORMATION
 *         AH = 38h
 * --DOS 2.x--
 *         AL = 00h get current-country info
 *         DS:DX -> buffer for returned info (see below)
 * Return: CF set on error
 *             AX = error code (02h)
 *         CF clear if successful
 *             AX = country code (MSDOS 2.11 only)
 *             buffer at DS:DX filled
 * --DOS 3+--
 *         AL = 00h for current country
 *         AL = 01h thru 0FEh for specific country with code <255
 *         AL = 0FFh for specific country with code >= 255
 *            BX = 16-bit country code
 *         DS:DX -> buffer for returned info (see below)
 * Return: CF set on error
 *             AX = error code (02h)
 *         CF clear if successful
 *             BX = country code
 *             DS:DX buffer filled
 * SeeAlso: AH=65h,INT 10/AX=5001h,INT 2F/AX=110Ch,INT 2F/AX=1404h
 * 
 * Format of PCDOS 2.x country info:
 * Offset  Size    Description
 *  00h    WORD    date format  0 = USA    mm dd yy
 *                              1 = Europe dd mm yy
 *                              2 = Japan  yy mm dd
 *  02h    BYTE    currency symbol
 *  03h    BYTE    00h
 *  04h    BYTE    thousands separator char
 *  05h    BYTE    00h
 *  06h    BYTE    decimal separator char
 *  07h    BYTE    00h
 *  08h 24 BYTEs   reserved
 * 
 * Format of MSDOS 2.x,DOS 3+ country info:
 * Offset  Size    Description
 *  00h    WORD    date format (see above)
 *  02h  5 BYTEs   ASCIZ currency symbol string
 *  07h  2 BYTEs   ASCIZ thousands separator
 *  09h  2 BYTEs   ASCIZ decimal separator
 *  0Bh  2 BYTEs   ASCIZ date separator
 *  0Dh  2 BYTEs   ASCIZ time separator
 *  0Fh    BYTE    currency format
 *                 bit 2 = set if currency symbol replaces decimal point
 *                 bit 1 = number of spaces between value and currency symbol
 *                 bit 0 = 0 if currency symbol precedes value
 *                         1 if currency symbol follows value
 *  10h    BYTE    number of digits after decimal in currency
 *  11h    BYTE    time format
 *                 bit 0 = 0 if 12-hour clock
 *                         1 if 24-hour clock
 *  12h    DWORD   address of case map routine
 *                 (FAR CALL, AL = character to map to upper case [>= 80h])
 *  16h  2 BYTEs   ASCIZ data-list separator
 *  18h 10 BYTEs   reserved
 * 
 * Values for country code:
 *  001h   United States
 *  002h   Canadian-French
 *  003h   Latin America
 *  01Fh   Netherlands
 *  020h   Belgium
 *  021h   France
 *  022h   Spain
 *  024h   Hungary (not supported by DR-DOS 5.0)
 *  026h   Yugoslavia (not supported by DR-DOS 5.0)
 *  027h   Italy
 *  029h   Switzerland
 *  02Ah   Czechoslovakia (not supported by DR-DOS 5.0)
 *  02Bh   Austria (DR-DOS 5.0)
 *  02Ch   United Kingdom
 *  02Dh   Denmark
 *  02Eh   Sweden
 *  02Fh   Norway
 *  030h   Poland (not supported by DR-DOS 5.0)
 *  031h   Germany
 *  037h   Brazil (not supported by DR-DOS 5.0)
 *  03Dh   International English [Australia in DR-DOS 5.0]
 *  051h   Japan (DR-DOS 5.0)
 *  052h   Korea (DR-DOS 5.0)
 *  15Fh   Portugal
 *  166h   Finland
 *  311h   Middle East (DR-DOS 5.0)
 *  3CCh   Israel (DR-DOS 5.0)
 * ----------2138-------------------------------
 * INT 21 - DOS 3+ - SET COUNTRY CODE
 *         AH = 38h
 *         AL = 01h thru 0FEh for specific country with code <255
 *         AL = FFh for specific country with code >= 255
 *            BX = 16-bit country code
 *         DX = FFFFh
 * Return: CF set on error
 *             AX = error code (see AH=59h)
 *         CF clear if successful
 * Note:   not supported by OS/2
 * SeeAlso: INT 2F/AX=1403h
 * 
 */
void dos_country_info(void)
{
	
	if(regs.al != 0) {
		fprintf(stderr, "trying to get info for country %d\n", regs.al);
		flags.carry = TRUE;
	} else {
		BYTE *info = fast_ds + regs.dx;

		regs.bx = 1;	/* US? */
		set_dummy_address(info + 12);
		flags.carry = FALSE;
	}
}


/* 
 * INT 21 U - DOS 3.1+ internal - SERVER FUNCTION CALL
 *         AX = 5D00h
 *         DS:DX -> DOS parameter list (see below)
 *         DPL contains all register values for a call to INT 21h
 * Return: as appropriate for function being called
 * Notes:  does not check AH.  Out of range values will crash the system
 *         executes using specified computer ID and process ID
 *         sharing delay loops skipped
 *         a special sharing mode is enabled
 *         wildcards are enabled for DELETE (AH=41h) and RENAME (AH=56h)
 *         an extra file attribute parameter is enabled for OPEN (AH=3Dh),
 *           DELETE (AH=41h), and RENAME (AH=56h)
 *         functions which take filenames require canonical names (as returned
 *           by AH=60h); this is apparently to prevent multi-hop file forwarding
 * SeeAlso: AH=3Dh,AH=41h,AH=56h,AH=60h
 * 
 * Format of DOS parameter list:
 * Offset  Size    Description
 *  00h    WORD    AX 
 *  02h    WORD    BX
 *  04h    WORD    CX
 *  06h    WORD    DX
 *  08h    WORD    SI
 *  0Ah    WORD    DI
 *  0Ch    WORD    DS
 *  0Eh    WORD    ES
 *  10h    WORD    reserved (0)
 *  12h    WORD    computer ID (0 = current system)
 *  14h    WORD    process ID (PSP segment on specified computer)
 * 
 * INT 21 U - DOS 3.1+ internal - COMMIT ALL FILES FOR SPECIFIED COMPUTER/PROCESS
        AX = 5D01h
        DS:DX -> DOS parameter list (see AX=5D00h), only computer ID and
                process ID fields used
Return: CF set on error
            AX = error code (see AH=59h)
        CF clear if successful
^LNotes:        flushes buffers and updates directory entries for each file whic
h has
          been written to; if remote file, calls INT 2F/AX=1107h
        the computer ID and process ID are stored but ignored under DOS 3.3
SeeAlso: AH=0Dh,AH=68h,INT 2F/AX=1107h
-------------------------------------------------------------------------------
INT 21 U - DOS 3.1+ internal - SHARE.EXE - CLOSE FILE BY NAME
        AX = 5D02h
        DS:DX -> DOS parameter list (see AX=5D00h), only fields DX, DS,
                computer ID, and process ID used
        DPL's DS:DX -> ASCIZ name of file to close
Return: CF set on error
            AX = error code (see AH=59h)
        CF clear if successful
Notes:  error unless SHARE is loaded (calls [SysFileTable-28h]) (see AH=52h)
        name must be canonical fully-qualified, such as returned by AH=60h
SeeAlso: AX=5D03h,AX=5D04h,AH=3Eh,AH=60h
-------------------------------------------------------------------------------
INT 21 U - DOS 3.1+ internal - SHARE.EXE - CLOSE ALL FILES FOR GIVEN COMPUTER
        AX = 5D03h
        DS:DX -> DOS parameter list (see AX=5D00h), only computer ID used
Return: CF set on error
            AX = error code (see AH=59h)
        CF clear if successful
Note:   error unless SHARE is loaded (calls [SysFileTable-30h]) (see AH=52h)
SeeAlso: AX=5D02h,AX=5D04h
-------------------------------------------------------------------------------
INT 21 U - DOS 3.1+ internal - SHARE.EXE - CLOSE ALL FILES FOR GIVEN PROCESS
        AX = 5D04h
        DS:DX -> DOS parameter list (see AX=5D00h), only computer ID and
                process ID fields used
 */
void dos_internal_server_function(void)
{
	fprintf(stderr, __FUNCTION__ " with function %d\n", regs.al);
}


static void get_boot_drive(void)
{
}

static void dr_dos_version_number()
{
}

void dos_extended_break_checking(void)
{
	printf(__FUNCTION__ " minor code %x\n", regs.al);
}


/* 
INT 21 - DOS 1+ - GET STDIN STATUS
        AH = 0Bh
Return: AL = 00h if no character available
           = FFh if character is available
Notes:  ^C/^Break checked, and INT 23 called if pressed
        standard input is always the keyboard under DOS 1.x, but may be
          redirected under DOS 2+
        if the interim console flag is set (see AX=6301h), this function
          returns AL=FFh if a partially-formed double-byte character is
          available
*/
void get_stdin_status(void)
{
	regs.al = 0; /* FIXME */
}

/*
 *----------2102-------------------------------
 *INT 21 - DOS 1+ - WRITE CHARACTER TO STANDARD OUTPUT
 *        AH = 02h
 *        DL = character to write
 *Return: AL = last character output (despite the official docs which state
 *                nothing is returned) (at least DOS 3.3-5.0)
 *Notes:  ^C/^Break are checked, and INT 23 executed if pressed
 *        standard output is always the screen under DOS 1.x, but may be
 *          redirected under DOS 2+
 *        the last character output will be the character in DL unless DL=09h
 *          on entry, in which case AL=20h as tabs are expanded to blanks
 *SeeAlso: AH=06h,AH=09h
 */
void write_character_to_stdout(void)
{
	write(1, &regs.dl, 1);
	regs.al = regs.dl;
}

get_extended_error_info(void)
{
	printf(__FUNCTION__ "with error code %x\n", regs.ax);
}

/*
 * INT 21 - DOS 2+ - GET FREE DISK SPACE
 *         AH = 36h
 *         DL = drive number (00h = default, 01h = A:, etc)
 * Return: AX = FFFFh if invalid drive
 *         else
 *             AX = sectors per cluster
 *             BX = number of free clusters
 *             CX = bytes per sector
 *             DX = total clusters on drive
 * Notes:  free space on drive in bytes is AX * BX * CX
 *         total space on drive in bytes is AX * CX * DX
 *         "lost clusters" are considered to be in use
 *         according to Dave Williams' MSDOS reference, the value in DX is
 *           incorrect for non-default drives after ASSIGN is run
 */
void get_dos_free_disk_space(void)
{
	printf("get free space for drive %d\n", regs.dl);
}


/*
INT 21 - DOS 1+ - GET CURRENT DEFAULT DRIVE
        AH = 19h
Return: AL = drive (00h = A:, 01h = B:, etc)
*/
void get_current_default_drive(void)
{
        regs.al = current_default_drive;  
}


/*
 *INT 21 - DOS 1+ - SELECT DEFAULT DRIVE
        AH = 0Eh
        DL = new default drive (00h = A:, 01h = B:, etc)
Return: AL = number of potentially valid drive letters
Notes:  under Novell NetWare, the return value is always 32, the number of
          drives that NetWare supports
        under DOS 3+, the return value is the greatest of 5, the value of
          LASTDRIVE= in CONFIG.SYS, and the number of drives actually present
        on a DOS 1.x/2.x single-floppy system, AL returns 2 since the floppy
          may be accessed as either A: or B:
        otherwise, the return value is the highest drive actually present
        DOS 1.x supports a maximum of 16 drives, 2.x a maximum of 63 drives,
          and 3+ a maximum of 26 drives
 */
void dos_select_default_drive(void)
{
	current_default_drive = regs.dl;
}
 

/* 
 * INT 21 - DOS 2+ - "CWD" - GET CURRENT DIRECTORY
 *         AH = 47h
 *         DL = drive number (00h = default, 01h = A:, etc)
 *         DS:SI -> 64-byte buffer for ASCIZ pathname
 * Return: CF clear if successful
 *             AX = 0100h (undocumented)
 *         CF set on error
 *             AX = error code (0Fh) (see AH=59h)
 * ^LNotes:        the returned path does not include a drive or the initial backsl
 * ash
 *         many Microsoft products for Windows rely on AX being 0100h on success
 *         under the FlashTek X-32 DOS extender, the buffer pointer is in DS:ESI
 * SeeAlso: AH=19h,AH=3Bh,INT 15/AX=DE25h
 */
void dos_getcwd(void)
{
	char *cp;


	cp = dos_getcwd_on_specific_drive(regs.dl);

	if(cp) {
		int size;
		
		size = strlen(cp) + 1;
		assert(size < 64);
		strcpy(fast_ds + regs.si, cp);
		regs.ax = 0x100;
		flags.carry = FALSE;		
	} else flags.carry = TRUE;;		
}


/* 
 * INT 21 - DOS 2+ - "CHDIR" - SET CURRENT DIRECTORY
 *         AH = 3Bh
 *         DS:DX -> ASCIZ pathname to become current directory (max 64 bytes)
 * Return: CF clear if successful
 *             AX destroyed
 *         CF set on error
 *             AX = error code (03h) (see AH=59h)
 * Notes:  if new directory name includes a drive letter, the default drive is
 *           not changed, only the current directory on that drive
 *         changing the current directory also changes the directory in which
 */ 
void dos_chdir(void)
{
	char *name;
	char *dos_name;

	dos_name = fast_ds + +regs.dx;

	if(!*dos_name) {
		/* fail */
		flags.carry = FALSE;
		return;
	}

	if(*(dos_name + 1) == ':') {
		printf("Can't support : yet\n");
		flags.carry = FALSE;
		return;
	}

	name = convert_dos_asciiz_to_unix_name(fast_ds + regs.dx);
	if(name) {
		int result;

		result = chdir(name);
		append_unix_name_to_current_drive(name);
		if(result == 0)
			flags.carry = FALSE;
		else	goto bad;
		return;
	} 

bad:
	fprintf(stderr, "tried dos_chdir to %s and couldn't\n", fast_ds + regs.dx);

	flags.carry = TRUE;
		
}

 

/* 
 * 
 * INT 21 - DOS 1+ - SET DISK TRANSFER AREA ADDRESS
 *         AH = 1Ah
 *         DS:DX -> Disk Transfer Area (DTA)
 * Notes:  the DTA is set to PSP:0080h when a program is started
 */
void dos_set_dta_address(void)
{

        current_task->dta  =  fast_ds + regs.dx;
	current_task->dta_seg = regs.ds;
	current_task->dta_offset = regs.dx;
}

/*
INT 21 - DOS 2+ - GET DISK TRANSFER AREA ADDRESS
        AH = 2Fh
Return: ES:BX -> current DTA
*/
void dos_get_dta_address(void)
{
	regs.es = current_task->dta_seg;
	regs.bx = current_task->dta_offset;
}

