/* 
 * Dos search file utilities
 *   Copyright (C) 1992  Martin Leisner  leisner@eso.mc.xerox.com 
 * 
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License
 *      as published by the Free Software Foundation; either version 2
 *      of the License, or (at your option) any later version.
 * 
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 * 
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include <ctype.h>
#include <dirent.h>

#include "misc.h"
#include "8086regs.h"
#include "memory.h"
#include "task.h"
#include "dosprotos.h"

#define DOS_RESERVED 21
#define ATTRIBUTE	(DOS_RESERVED )
#define FILE_TIME	(ATTRIBUTE + 1)	
#define FILE_DATE	(FILE_TIME + 2)	
#define FILE_SIZE	(FILE_DATE + 2)	
#define FILE_NAME	(FILE_SIZE + 4)	 

static char *get_name_ptr(void)
{
	return current_task->dta + FILE_NAME;
}

static char *get_attribute_ptr(void)
{
	return current_task->dta + ATTRIBUTE;
}

static char *get_time_ptr(void)
{
	return current_task->dta + FILE_TIME;
}

static char *get_date_ptr(void)
{
	return current_task->dta + FILE_DATE;
}


static char *get_size_ptr(void)
{
	return current_task->dta + FILE_SIZE ;
}

/* get the date, time and size and fill it in.
 * Return TRUE if we can stat, FALSE if we can't
 */
static BOOLEAN fill_in_rest(const char *filename)
{
	char fullname[256];
	struct stat statbuf;
	int result;
	char *cp;	/* pointer for date, time, size? */
	DOUBLE_WORD	dos_time;

	sprintf(fullname, "%s/%s", current_task->path_for_dir, filename);	
	
	/* use lstat? */
	result = stat(fullname, &statbuf);

	/* stat can fail if it is a symlink and the link 
         * isn't there */
	
	if(result < 0)
		return FALSE;
#if 0 
	printf("stating %s, result = %d ", fullname, result);
	printf("size = %d\n", statbuf.st_size);
#endif


	
	

	cp = get_size_ptr();

#if 1
	write_pc_word(cp, statbuf.st_size & 0xffff);	/* LSW */
	write_pc_word(cp + 2, statbuf.st_size >> 16);	/* MSW */	
#else
	write_pc_word(cp + 2, 0x100);	/* LSW */
	write_pc_word(cp, 0x200);	/* MSW */	

#endif
	dos_time = convert_unix_time(statbuf.st_mtime);		

	cp = get_time_ptr();
	write_pc_word(cp, (WORD) (dos_time & 0xffff));
	
	cp = get_date_ptr();
	write_pc_word(cp, (WORD) (dos_time >> 16));
		
	regs.cx = 0;
	return TRUE;

	}

	/* return the directory found, and slot the file name into file
	 * (should be at least 13 characters)
	 */ 
	static char *find_first_directory_and_file(char *name)
	{
		static char path[64];
		char *last;
		char *cp;

		/* check for absolute path (beginning with A-Za-z:) */
		if(*(name + 2) == ':') {
			char *dos_path;
			if(isupper(*name))
				dos_path = get_drive_path_for_unix(*name - 'A'); 
			else if(islower(*name))
				dos_path = get_drive_path_for_unix(*name - 'a');
			else {
				printf("Invalid path = %s\n", name);
				abort();
			}
			assert(dos_path);
			strcpy(path, dos_path);
		} else strcpy(path, ".");
		
		last = rindex(name, '/');
		if(last) {
			if(rindex(name, '\\')) { 
				printf("mixing up / and \\");
				abort();
			} 
			last++;
		} else {
			last = rindex(name, '\\');
			if(last)
				last++;
		}
		if(last) {
			for(cp = path; name < last - 1; name++, cp++) {
				if(isupper(*name))
					*cp = tolower(*name);
				else if(*name == '\\')
					*cp = '/';
				else *cp = *name;	/* check for ascii? */
			}
			
		} else last = name; 

		strncpy(current_task->find_pattern, last, sizeof(current_task->find_pattern) - 1);
		return path;		

		
	}


	static BOOLEAN compare_files(char *mask, char *name)
	{
		if(!strcmp(name, ".") || !strcmp(name, ".."))
			return FALSE;

		while(*mask && *name) {
			/* match any character */
			if(*mask == '?') {
				mask++;
				if(*name != '.')
					name++;
				continue;
			}

			if(*name == '*') {
				printf("can't have name with *\n");
				return FALSE;
			}

			/* wildcard */
			if(*mask == '*') {
				while(*name && *name != '.')
					name++;
				mask++;
				continue;
			}
			
			/* match */             
			if(toupper(*mask) == toupper(*name)) {
				mask++;
				name++;
				continue;
			} else return FALSE;        /* no match */

		}

		if(!*mask && !*name)
			return TRUE;       /* matched */
		else    return FALSE;

	}

	/* sets flags.carry TRUE if no file is found, otherwise FALSE.
	 * slots into the approrpiate stuff
	 */
	static void get_next_find(void)
	{
		struct dirent *file;
		int i;
		/* the filename mask where looking for */
		char *mask;
		/* the name actually found */
		char *name;

		while(1) {
			file = readdir(current_task->find_first);
			if(NULL == file) {
				/* end of list, close dir and indicate none found */
				closedir(current_task->find_first);
				current_task->find_first = NULL;
				current_task->ff_strategy = FF_NONE;
				flags.carry = TRUE;
				return;
			}
			if(file->d_name[0] == '.')
				continue; 
			if(compare_files(current_task->find_pattern, 
					 file->d_name) == TRUE) {
				int i;
				char *name;
				BOOLEAN fill_in_status;
				
				name = get_name_ptr();

				/* copy name and convert to upper case */
				for(i = 0; i < 8 + 3 + 1; i++) {
					if(file->d_name[i] == '\0')
						break;
					*name++ = toupper(file->d_name[i]);		
				}
				*name = '\0';
				fill_in_status = fill_in_rest(file->d_name);
				if(fill_in_status == FALSE)
					continue;	/* probably a broken symlink */
			
				/* have match */
				flags.carry = FALSE;
				return;
			}	
		}			
				
				
	}

	/*
	 * INT 21 - DOS 2+ - "FINDFIRST" - FIND FIRST MATCHING FILE
	 *         AH = 4Eh
	 *         AL = special flag for use by APPEND (see note below)
	 *         CX = file attribute mask (see AX=4301h) (bits 0 and 5 ignored)
	 *         DS:DX -> ASCIZ file specification (may include path and wildcards)
	 * Return: CF clear if successful
	 *             [DTA] = FindFirst data block (see below)
	 *         CF set on error
	 *             AX = error code (02h,03h,12h) (see AH=59h)
	 * 
	 * CX flags:
	 * bit 7  shareable (Novell NetWare)
	 bit 6  unused
	 bit 5  archive
	 bit 4  directory
	 bit 3  volume label
		execute-only (Novell NetWare)
	 bit 2  system
	 bit 1  hidden
	 bit 0  read-only

	Format of FindFirst data block:
	Offset  Size    Description
	---PC-DOS 3.10, PC-DOS 4.01, MS-DOS 3.2/3.3/5.0---
	 00h    BYTE    drive letter (bits 0-6), remote if bit 7 set
	 01h 11 BYTEs   search template
	 0Ch    BYTE    search attributes
	---DOS 2.x (and some DOS 3.x???)---
	 00h    BYTE    search attributes
	 01h    BYTE    drive letter
	 02h 11 BYTEs   search template
	---WILDUNIX.COM---
	 00h 12 BYTEs   15-character wildcard search pattern and drive letter (packed)
	 0Ch    BYTE    search attributes
	---DOS 2.x and most 3.x---
	 0Dh    WORD    entry count within directory
	 0Fh    DWORD   pointer to DTA???
	 13h    WORD    cluster number of start of parent directory
	---PC-DOS 4.01, MS-DOS 3.2/3.3/5.0---
	 0Dh    WORD    entry count within directory
	 0Fh    WORD    cluster number of start of parent directory
	 11h  4 BYTEs   reserved
	---all versions, documented fields---
	 15h    BYTE    attribute of file found
	 16h    WORD    file time
			    bits 11-15: hour
			    bits 5-10:  minute
			    bits 0-4:   seconds/2
	 18h    WORD    file date
			    bits 9-15:  year-1980
			    bits 5-8:   month
			    bits 0-4:   day
	 1Ah    DWORD   file size
 1Eh 13 BYTEs   ASCIZ filename+extension

 *
 */
void dos_find_first_file(void)
{
        char *directory;

        if(current_task->find_first)
                closedir(current_task->find_first);
        directory = find_first_directory_and_file(fast_ds + regs.dx);
        if(directory)  {
                /* have a directory */
                current_task->find_first = opendir(directory);
		if(current_task->find_first == NULL) {
			flags.carry = TRUE;
			return;
		}
		strcpy(current_task->path_for_dir, directory);
                current_task->find_first_mask = regs.cl;                
		current_task->ff_strategy = FF_DTA;		
		get_next_find() ;
		return;
        }
        flags.carry = TRUE;

}

/*
 * INT 21 - DOS 2+ - "FINDNEXT" - FIND NEXT MATCHING FILE
 *         AH = 4Fh
 *         [DTA] = data block from previous FindFirst or FindNext call
 * Return: CF clear if successful
 *         CF set on error
 *             AX = error code (12h) (see AH=59h)
 */
void dos_find_next_file(void)
{
	assert(current_task->ff_strategy == FF_DTA);
	get_next_find();
}

