/*
 *   This is msdos, an msdos/ibmpc emulator.
 *   Copyright (C) 1992  Martin Leisner  leisner@eso.mc.xerox.com 
 * 
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License
 *      as published by the Free Software Foundation; either version 2
 *      of the License, or (at your option) any later version.
 * 
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 * 
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <stdio.h>
#include <strings.h>
#include <signal.h>

#include "misc.h"
#include "8086regs.h"

char *exe_file;
static char *progname;


typedef struct { 
        /* debug flag -- if 0, end of list  */
        char c;
        /* where the flag actually is */
        BOOLEAN *where;
        /* an ascii description */
        char *description;
} DEBUG_FLAGS;
	

static DEBUG_FLAGS debug_flags[] = {
 	{	'i', &dos_show_io, "show dos io operations"	 },
	{ }
};



static BOOLEAN test_debug(char c, DEBUG_FLAGS *debug_flags)
{
	while(debug_flags->c != 0) {
		if(debug_flags->c == c) {
			*(debug_flags->where) = (*(debug_flags->where) == TRUE) ? 
				FALSE : TRUE;
			return TRUE;
		}
		debug_flags++;
			 
	}
	return FALSE;
}

static char *set_debug_flags(char *string)
{
       	BOOLEAN status;
 
        while(*string) {
                status = test_debug(*string,  &debug_flags);
        
                if(status == FALSE) 
                        return string;
                string++;
        }
	return NULL;

}

static void show_debug_flags(void)
{
	DEBUG_FLAGS *p;
	fprintf(stderr, "\tDebug flags are:\n");
	for(p = &debug_flags; p->c != 0; p++) {
                fprintf(stderr, "\t\t%c\t", p->c);
                fprintf(stderr, "%s", p->description ? p->description :
                        "not defined");
                if(p->where)
                        fprintf(stderr, "\tcurrently %s\n", (*p->where) ? "TRUE" : "FALSE");
                else fprintf(stderr, "\tNot pointing at anything\n");
	}	
}
	
/* bring the file into ram */
load_file(void)
{
	FILE *fp;
	int c;
	char *p = &pc_ram[boot_regs.cs << 4];
	char *cp;

	fp = fopen(exe_file, "r");
	if(!fp) {
		fprintf(stderr, "can't open %s\n", exe_file);
		return 1;
	}
	
	cp = rindex(exe_file, '.');
	if(cp &&  !strcmp(cp, ".exe")) {
		if(fseek(fp, (long) 512, 0)) {
			perror("fseek failed");
			return;
		}
	}

	while((c = fgetc(fp)) != EOF) 
		*p++ = c;	
	fclose(fp);
	printf("Read %d bytes\n", p - (char *) &pc_ram[boot_regs.cs << 4]);
	return 0;	
}

/* assing a register in the form cc=0xdddd where cc is: cs, ds, es, ss, sp, ip 
 * return TRUE if it works, FALSE otherwise
 */
static BOOLEAN try_to_parse_arg(const char *arg)
{
	WORD_REG *preg = NULL;
	unsigned temp;
	int i;

	if(!strncmp(arg, "cs", 2)) 
		preg = &boot_regs.cs;
	else if(!strncmp(arg, "ds", 2))
		preg = &boot_regs.ds;
	else if(!strncmp(arg, "es", 2))
		preg = &boot_regs.es;
	else if(!strncmp(arg, "ss", 2))
		preg = &boot_regs.ss;
	else if(!strncmp(arg, "sp", 2))
		preg = &boot_regs.sp;

	if(!preg)
		return FALSE;	/* no match */
	
	i = sscanf(arg + 2, "=0x%x", &temp)	;
	if(i != 1)
		return FALSE;
	*preg = temp;
	return TRUE;	
}

/* get registers to boot with from a file called called filename.trc
 * Return FALSE for problems, otherwise TRUE.
 */
static BOOLEAN get_aztec_trc_registers(const char *filename)
{
	char trc_file[80];
	char buffer[100];
	FILE *trace;
	char *cp;
	int i;
	int sp, ss, ds, es, cs;
	int ip;	
	strcpy(trc_file, filename);
	strcat(trc_file, ".trc");	
	trace = fopen(trc_file, "r");
	if(trace == NULL) {
		fprintf(stderr, "Can't open %s\n", trc_file);
		return FALSE;
	}

	/* get sp from first line */
	fgets(buffer, sizeof(buffer), trace);
	cp = index(buffer, 's');
	if(strncmp(cp, "sp=", 3)) {
		fprintf(stderr, "Can't find sp in %s\n", buffer);
		return FALSE;
	}

	i =sscanf(cp+3, "%x", &sp);
	if(i != 1) {
		fprintf(stderr, "Can't find sp in %s\n", buffer);
		return FALSE;
	}
	
	/* get next line */
	fgets(buffer, sizeof(buffer), trace);
	i = sscanf(buffer, "cs=%x  ds=%x  ss=%x  es=%x  ip=%x", 
		&cs, &ds, &ss, &es, &ip);
	if(i != 5) {
		fprintf(stderr, "Can't convert %s\n", buffer);
		return FALSE;
	}
	boot_regs.sp = sp;
	boot_regs.cs = cs;
	boot_regs.ds = ds;
	boot_regs.es = es;
	boot_regs.ss = ss;
	/*boot_regs.ip = ip; */
	return TRUE;
}


static void segmentation_fault()
{
	fprintf(stderr, "caught segmentation fault\n");
	exit();
}

static void ignore_math(void)
{
	puts(__FUNCTION__);
	printf("Instruction %d: cs:ip = %x:%x\n", num_instructions, regs.cs, regs.ip);
}

static void usage(void)
{
	fprintf(stderr, "Run %s [-e environment ] [-dflags] [-s] progname [ progargs]\n", progname);
	fprintf(stderr, "\t-s is slow/break (normally fast)\n");
	show_debug_flags();
	exit(1);
}

main(int argc, char **argv) 
{	
	extern int optind;
	extern char *optarg;
	int c;
	char *dos_name;
	char *dos_options;
	BOOLEAN run_slow = FALSE;
	
	signal(SIGSEGV, segmentation_fault);
	signal(SIGFPE,  ignore_math);

	progname = argv[0];
#if 0
                if(setjmp(on_fatal_error)) 
                        abort();
                if(setjmp(after_halt)) 
                      	exit();
#endif

	printf("DOS EMULATOR PID = %d\n", getpid());


	dos_environment = getenv("DOSENVIRON");
	while(1) {
		c = getopt(argc, argv, "d:s");
		if(c == -1)
			break;
		switch(c) {
			case 's':
				run_slow = TRUE;
				break;
			case 'd':
				set_debug_flags(optarg);
				break;
			case 'e':
				dos_environment = optarg;
				break;
			default:
				usage();
		}
	}	

	if(optind >= argc)
		usage();

	dos_name = argv[optind++];

	if(optind < argc)
		dos_options = argv[optind];
	else	dos_options = NULL;

	try_to_read_symbol_table(dos_name);

	if(boot_msdos(dos_name, dos_options ) == FALSE) {
		fprintf(stderr, "can't boot msdos\n");
		exit(1);
	}
	if(rindex(argv[1], '/'))
		exe_file = rindex(dos_name, '/') + 1;
	else	exe_file = dos_name;

	boot_processor();


	if(run_slow == TRUE) {
		while(1) {
			num_instructions++;
			test_breakpoint();
			print_like_aztec();
			emulate_instruction();
		}
	} else
		while(1) {
			num_instructions++;
			emulate_instruction();
		}

} 


