/*
 *
 *    This is part of an msdos/ibmpc emulator.
 *   Copyright (C) 1992  Martin Leisner  leisner@eso.mc.xerox.com 
 * 
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License
 *      as published by the Free Software Foundation; either version 2
 *      of the License, or (at your option) any later version.
 * 
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 * 
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * 
 * Way to manage memory
 * $Log$
 */

#ifndef _MEMORY_H
#define _MEMORY_H


EXTERN BOOLEAN use_disk_get_environment INIT(FALSE);
/* the dos list points to blocks used in msdos memory managed 
 * the proc list points to the blocks used by a process (nominally 1)
 */


/* each block is either on the stack, in the free pool or allocated
 * to a process 
 */
typedef enum { FREE_BLOCK, STACK_BLOCK, ALLOCATED_BLOCK, UNKNOWN_BLOCK } TYPE_BLOCK;

/* values are the was int21/ah=0x58 defines them */
typedef enum { FIRST_FIT = 0, BEST_FIT = 1, LAST_FIT = 2 } MEM_ALLOC_STRATEGY;

EXTERN MEM_ALLOC_STRATEGY mem_alloc_strategy;

typedef struct memory_block MEM_BLOCK;


/* each memory block has this structure.
 * It is really one click bigger than asked for
 *
 * Format of memory control block (see also below):
Offset  Size    Description
 00h    BYTE    block type: 5Ah if last block in chain, otherwise 4Dh
 01h    WORD    PSP segment of owner or
                0000h if free
                0006h if DR-DOS XMS UMB
                0007h if DR-DOS excluded upper memory ("hole")
                0008h if belongs to DOS
                FFFAh if 386MAX UMB control block (see AX=4402h"386MAX")
                FFFDh if 386MAX locked-out memory
                FFFEh if 386MAX UMB (immediately follows its control block)
                FFFFh if 386MAX 6.01 device driver
 03h    WORD    size of memory block in paragraphs
 05h  3 BYTEs   unused by MS-DOS
                (386MAX) if locked-out block, region start/prev region end
 */
struct memory_block {
	/* doubly linked list for free list and dos allocated list.
 	 * End of list is signaled by a NULL */
	MEM_BLOCK *next;
	MEM_BLOCK  *prev;
	struct _task         *owner;	/* task which owns this */
	char *dos_memory;		/* where this points to in dos memory */
	/* this isn't necessary, but may be useful for debugging */
	TYPE_BLOCK	type;
	/* usable space (16 bytes for memory allocation at head) */
	int	size;		/* in clicks (16 byte units) --
				 * this is the total size of the memory 
				 * allocation, including a dos arena header.
				 */ 
	int	segment;	/* where the block sits */
}; 


#endif

