
#ifndef _TASK_H
#define _TASK_H

#include <dirent.h>
#include "memory.h"
/*
 *
 * generate our copy of a psp, which is close to dos.
 * we keep our own copy because we don't want to deal with the dos one,
 * but make it as accurate as we can 
 */

 
/* From Ralf Brown's interrupt list:
 *
 *format of PSP:
 *Offset  Size    Description
 * 00h  2 BYTEs   INT 20 instruction for CP/M CALL 0 program termination
 *                the CDh 20h here is often used as a signature for a valid PSP
 * 02h    WORD    segment of first byte beyond memory allocated to program
 * 04h    BYTE    unused filler
 * 05h    BYTE    CP/M CALL 5 service request (FAR JMP to 000C0h)
 *                BUG: (DOS 2+) PSPs created by INT 21/AH=4Bh point at 000BEh
 * 06h    WORD    CP/M compatibility--size of first segment for .COM files
 * 08h  2 BYTEs   remainder of FAR JMP at 05h
 * 0Ah    DWORD   stored INT 22 termination address
 * 0Eh    DWORD   stored INT 23 control-Break handler address
 * 12h    DWORD   DOS 1.1+ stored INT 24 critical error handler address
 * 16h    WORD    segment of parent PSP
 * 18h 20 BYTEs   DOS 2+ Job File Table, one byte per file handle, FFh = closed
 * 2Ch    WORD    DOS 2+ segment of environment for process
 * 2Eh    DWORD   DOS 2+ process's SS:SP on entry to last INT 21 call
 * 32h    WORD    DOS 3+ number of entries in JFT (default 20)
 * 34h    DWORD   DOS 3+ pointer to JFT (default PSP:0018h)
 * 38h    DWORD   DOS 3+ pointer to previous PSP (default FFFFFFFFh in 3.x)
 *                used by SHARE in DOS 3.3
 * 3Ch  4 BYTEs   unused by DOS versions <= 5.00
 *                reportedly used by Novell NetWare shell 3.x
 * 40h  2 BYTEs   DOS 5.0 version to return on INT 21/AH=30h
 * 42h  6 BYTEs   unused by DOS versions <= 5.00
 * 48h    BYTE    (MSWindows3) bit 0 set if non-Windows application (WINOLDAP)
 * 49h  7 BYTEs   unused by DOS versions <= 5.00
 * 50h  3 BYTEs   DOS 2+ service request (INT 21/RETF instructions)
 * 53h  9 BYTEs   unused in DOS versions <= 5.00
 * 5Ch 16 BYTEs   first default FCB, filled in from first commandline argument
 *                overwrites second FCB if opened
 * 6Ch 16 BYTEs   second default FCB, filled in from second commandline argument
 *                overwrites beginning of commandline if opened
 * 7Ch  4 TEs   unused
 * 80h 128 BYTEs  commandline / default DTA
 *                command tail is BYTE for length of tail, N BYTEs for the tail,
 *                followed by a BYTE containing 0Dh 
 */

typedef struct _task TASK;
struct _task  {
	TASK *parent;
	MEM_BLOCK *psp_memory;	/* points to memory block psp is at head of */
	char *psp;	/* where the psp is in dos memory for this task */
	char *jft;	/* point to pc ram job file table -- 20 bytes of
			 * characters which map into another table
		         * of handles
			 */
	WORD dta_seg;	/* segment of dta */
	WORD dta_offset; /* offset of dta */
	char *dta;	/* where the dta for this task is in dos memory */
	REGS_8086 saved_regs;  /* save the regs when switching processes  -- this is 
				* because we don't do real interrupts */
	FLAGS saved_flags;
	/* where psp is located */
	WORD   psp_click;	
	BOOLEAN interrupt_23;
	WORD int23_seg;
	WORD int23_off;
	BYTE return_code;
	DIR *find_first;	/* for find first routines */
	char path_for_dir[256];	/* should use some MAXPATH */
	BYTE find_first_mask;	/* mask for INT21:0x4f calls */
	enum { FF_FCB, 	
	       FF_DTA,	/* interrupt 4f calls */
		FF_NONE } ff_strategy;

	char find_pattern[13];
	TASK *magic;
	
}; 

enum { JFT_SIZE = 20 };

#define FCB1(task)	(task + 0x5c)
#define FCB2(task)	(task + 0x6c)
#define SEG_ENV(task)	(task + 0x2c)

/* we use a linked list to go through the task -- current_task is the newest
 * task, the last task on the list points to null (through parent) and
 * we have to execute tasks sequentially
 */
EXTERN TASK *current_task;

EXTERN char current_default_drive INIT('D'- 'A');
#endif

